from typing import Optional

from django.contrib import admin
from django.db.models import Count
from eveuniverse.models import EveEntity

from .core import ContactType
from .models import ContactSet, RequestLogEntry, StandingRequest, StandingRevocation


class AbstractStandingsRequestAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "_contact_type_str",
        "_contact_name",
        "_user",
        "request_date",
        "action_by",
        "action_date",
        "is_effective",
        "effective_date",
    )
    list_filter = ("is_effective",)
    list_select_related = True
    ordering = ("-id",)

    def _contact_name(self, obj):
        return EveEntity.objects.resolve_name(obj.contact_id)

    @admin.display(description="contact type")
    def _contact_type_str(self, obj):
        if obj.contact_type_id in ContactType.character_ids:
            return "Character"
        elif obj.contact_type_id in ContactType.corporation_ids:
            return "Corporation"
        return "(undefined)"

    def _user(self, obj):
        try:
            return obj.user
        except AttributeError:
            return None

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(StandingRequest)
class StandingsRequestAdmin(AbstractStandingsRequestAdmin):
    pass


@admin.register(StandingRevocation)
class StandingsRevocationAdmin(AbstractStandingsRequestAdmin):
    pass


@admin.register(ContactSet)
class ContactSetAdmin(admin.ModelAdmin):
    change_list_template = "admin/standingsrequests/contactset/change_list.html"
    list_display = ("date", "_contacts_count")
    list_display_links = None
    ordering = ("-date",)

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.annotate(contacts_count=Count("contacts"))

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False

    def _contacts_count(self, obj):
        return obj.contacts_count


@admin.register(RequestLogEntry)
class RequestLogEntryAdmin(admin.ModelAdmin):
    list_display = (
        "created_at",
        "request_type",
        "_requested_for",
        "_requested_by",
        "requested_at",
        "_reason",
        "action",
        "_action_by",
    )
    list_display_links = None
    list_filter = (
        "request_type",
        ("action_by", admin.RelatedOnlyFieldListFilter),
        ("requested_by", admin.RelatedOnlyFieldListFilter),
        "created_at",
        "reason",
    )
    ordering = ("-created_at",)

    @admin.display(ordering="action_by")
    def _action_by(self, obj) -> str:
        return "SYSTEM" if obj.action_by is None else obj.action_by.html()

    @admin.display(ordering="requested_by")
    def _requested_by(self, obj) -> str:
        return obj.requested_by.html()

    @admin.display(ordering="requested_for")
    def _requested_for(self, obj) -> str:
        return obj.requested_for.html()

    @admin.display(ordering="reason")
    def _reason(self, obj) -> Optional[str]:
        reason_obj = StandingRequest.Reason(obj.reason)
        return None if reason_obj is StandingRequest.Reason.NONE else reason_obj.label

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related(
            "action_by__character",
            "action_by__corporation",
            "requested_by__character",
            "requested_by__corporation",
            "requested_for__character",
            "requested_for__corporation",
        )

    def has_change_permission(self, *args, **kwargs):
        return False

    def has_add_permission(self, *args, **kwargs):
        return False

    # TODO: enable for production
    # def has_delete_permission(self, *args, **kwargs) -> bool:
    #     return False



# flake8: noqa

from .views_1 import (
    admin_changeset_update_now,
    create_requests,
    index_view,
    remove_character_standing,
    remove_corp_standing,
    request_character_standing,
    request_characters,
    request_corp_standing,
    request_corporations,
    view_auth_page,
    view_requester_add_scopes,
)
from .views_2 import (
    _compose_standing_requests_data,
    download_pilot_standings,
    manage_get_requests_json,
    manage_get_revocations_json,
    manage_requests_write,
    manage_revocations_write,
    manage_standings,
    view_active_requests,
    view_groups_standings,
    view_groups_standings_json,
    view_pilots_standings,
    view_pilots_standings_json,
    view_requests_json,
)



# Generated by Django 3.1.10 on 2021-05-13 19:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("standingsrequests", "0007_big_performance_overhaul"),
    ]

    operations = [
        migrations.AddField(
            model_name="standingrevocation",
            name="reason",
            field=models.CharField(
                choices=[
                    ("NO", "None recorded"),
                    ("OR", "Requested by character owner"),
                    ("LP", "Character owner has lost permission"),
                    ("CT", "Not all corp tokens are recorded in Auth."),
                    ("RG", "Standing has been revoked in game"),
                ],
                default="NO",
                max_length=2,
            ),
        ),
    ]



from allianceauth.eveonline.evelinks import eveimageserver

from ..models import CharacterAffiliation


class EveCharacterHelper:
    """
    Mimics Alliance Auths EveCharacter with internal standingstool data instead
    """

    # Not implemented
    corporation_ticker = None

    user = None

    def __init__(self, character_id):
        self.character_id = int(character_id)
        self.alliance_name = None
        try:
            assoc = CharacterAffiliation.objects.select_related(
                "character", "corporation", "alliance"
            ).get(character_id=self.character_id)
        except CharacterAffiliation.DoesNotExist:
            assoc = None
            self.corporation_id = None
            self.corporation_name = None
            self.alliance_id = None
        else:
            self.corporation_id = assoc.corporation_id
            self.alliance_id = assoc.alliance_id

        self.character_name = (
            assoc.character.name if assoc and assoc.character else None
        )
        self.corporation_name = (
            assoc.corporation.name if assoc and assoc.corporation else None
        )
        self.alliance_name = assoc.alliance.name if assoc and assoc.alliance else None

    def portrait_url(self, size: int = eveimageserver._DEFAULT_IMAGE_SIZE) -> str:
        return eveimageserver.character_portrait_url(self.character_id, size)



from django.urls import path

from . import views

app_name = "standingsrequests"

urlpatterns = [
    path("", views.index_view, name="index"),
    path("create_requests", views.create_requests, name="create_requests"),
    path("request_characters", views.request_characters, name="request_characters"),
    path(
        "request_corporations", views.request_corporations, name="request_corporations"
    ),
    path(
        "request_character_standing/<int:character_id>/",
        views.request_character_standing,
        name="request_character_standing",
    ),
    path(
        "remove_character_standing/<int:character_id>/",
        views.remove_character_standing,
        name="remove_character_standing",
    ),
    path(
        "request_corp_standing/<int:corporation_id>/",
        views.request_corp_standing,
        name="request_corp_standing",
    ),
    path(
        "remove_corp_standing/<int:corporation_id>/",
        views.remove_corp_standing,
        name="remove_corp_standing",
    ),
    path("view/pilots/", views.view_pilots_standings, name="view_pilots"),
    path(
        "view/pilots/json/", views.view_pilots_standings_json, name="view_pilots_json"
    ),
    path(
        "view/pilots/download/",
        views.download_pilot_standings,
        name="download_pilots",
    ),
    path("view/corps/", views.view_groups_standings, name="view_groups"),
    path("view/corps/json", views.view_groups_standings_json, name="view_groups_json"),
    path("manage/", views.manage_standings, name="manage"),
    path(
        "manage/requests/",
        views.manage_get_requests_json,
        name="manage_get_requests_json",
    ),
    # Should always follow the path of the GET path above
    path(
        "manage/requests/<int:contact_id>/",
        views.manage_requests_write,
        name="manage_requests_write",
    ),
    path(
        "manage/revocations/",
        views.manage_get_revocations_json,
        name="manage_get_revocations_json",
    ),
    path(
        "manage/revocations/<int:contact_id>/",
        views.manage_revocations_write,
        name="manage_revocations_write",
    ),
    path("view/requests/", views.view_active_requests, name="view_requests"),
    path("view/requests/json/", views.view_requests_json, name="view_requests_json"),
    path("manage/setuptoken/", views.view_auth_page, name="view_auth_page"),
    path(
        "requester_add_scopes/",
        views.view_requester_add_scopes,
        name="view_requester_add_scopes",
    ),
    path(
        "admin_changeset_update_now/",
        views.admin_changeset_update_now,
        name="admin_changeset_update_now",
    ),
]



from typing import Optional, Tuple

from bravado.exception import HTTPError

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction
from django.db.models import Case, Q, Value, When
from django.utils.translation import gettext_lazy as _
from esi.models import Token
from eveuniverse.models import EveEntity
from eveuniverse.tasks import update_unresolved_eve_entities

from allianceauth.eveonline.models import EveCharacter
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.helpers import chunks
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import SR_NOTIFICATIONS_ENABLED
from .constants import CreateCharacterRequestError, OperationMode
from .core import BaseConfig, ContactType
from .providers import esi

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class ContactSetManager(models.Manager):
    def create_new_from_api(self) -> object:
        """fetches contacts with standings for configured alliance
        or corporation from ESI and stores them as newly created ContactSet

        Returns new ContactSet on success, else None
        """
        owner_character = BaseConfig.owner_character()
        token = (
            Token.objects.filter(character_id=owner_character.character_id)
            .require_scopes(self.model.required_esi_scope())
            .require_valid()
            .first()
        )
        if not token:
            logger.warning("Token for standing char could not be found")
            return None
        try:
            contacts_wrap = _ContactsWrapper(token, owner_character)
        except HTTPError as ex:
            logger.exception(
                "APIError occurred while trying to query api server: %s", ex
            )
            return None

        with transaction.atomic():
            contacts_set = self.create()
            self._add_labels_from_api(contacts_set, contacts_wrap.labels)
            self._add_contacts_from_api(contacts_set, contacts_wrap.contacts)

        return contacts_set

    def _add_labels_from_api(self, contact_set, labels):
        """Add the list of labels to the given ContactSet

        contact_set: ContactSet instance
        labels: Label dictionary
        """
        from .models import ContactLabel

        contact_labels = [
            ContactLabel(label_id=label.id, name=label.name, contact_set=contact_set)
            for label in labels
        ]
        ContactLabel.objects.bulk_create(contact_labels, ignore_conflicts=True)

    def _add_contacts_from_api(self, contact_set, contacts):
        """Add all contacts to the given ContactSet
        Labels _MUST_ be added before adding contacts

        :param contact_set: Django ContactSet to add contacts to
        :param contacts: List of _ContactsWrapper.Contact to add
        """
        from .models import Contact

        for contact in contacts:
            eve_entity, _ = EveEntity.objects.get_or_create_esi(id=contact.id)
            obj = Contact.objects.create(
                contact_set=contact_set,
                eve_entity=eve_entity,
                standing=contact.standing,
            )
            flat_labels = [label.id for label in contact.labels]
            labels = contact_set.labels.filter(label_id__in=flat_labels)
            obj.labels.add(*labels)


class _ContactsWrapper:
    """Converts raw contacts and contact labels data from ESI into an object"""

    class Label:
        def __init__(self, json):
            self.id = json["label_id"]
            self.name = json["label_name"]

        def __str__(self) -> str:
            return str(self.name)

        def __repr__(self) -> str:
            return str(self)

    class Contact:
        def __init__(self, json, labels, names_info):
            self.id = json["contact_id"]
            self.name = names_info[self.id] if self.id in names_info else ""
            self.standing = json["standing"]
            self.in_watchlist = json["in_watchlist"] if "in_watchlist" in json else None
            self.label_ids = (
                json["label_ids"]
                if "label_ids" in json and json["label_ids"] is not None
                else []
            )
            # list of labels
            self.labels = [label for label in labels if label.id in self.label_ids]

        def __str__(self) -> str:
            return str(self.name)

        def __repr__(self):
            return str(self)

    def __init__(self, token, owner_character):
        self.contacts = []
        self.labels = []

        if BaseConfig.operation_mode is OperationMode.ALLIANCE:
            if not owner_character.alliance_id:
                raise RuntimeError(
                    "{owner_character}: owner character is not a member of an alliance"
                )
            labels = esi.client.Contacts.get_alliances_alliance_id_contacts_labels(
                alliance_id=owner_character.alliance_id,
                token=token.valid_access_token(),
            ).results()
            self.labels = [self.Label(label) for label in labels]
            contacts = esi.client.Contacts.get_alliances_alliance_id_contacts(
                alliance_id=owner_character.alliance_id,
                token=token.valid_access_token(),
            ).results()

        elif BaseConfig.operation_mode is OperationMode.CORPORATON:
            labels = (
                esi.client.Contacts.get_corporations_corporation_id_contacts_labels(
                    corporation_id=owner_character.corporation_id,
                    token=token.valid_access_token(),
                ).results()
            )
            self.labels = [self.Label(label) for label in labels]
            contacts = esi.client.Contacts.get_corporations_corporation_id_contacts(
                corporation_id=owner_character.corporation_id,
                token=token.valid_access_token(),
            ).results()
        else:
            raise NotImplementedError()

        logger.debug("Got %d contacts in total", len(contacts))
        entity_ids = [contact["contact_id"] for contact in contacts]
        resolver = EveEntity.objects.bulk_resolve_names(entity_ids)
        self.contacts = [
            self.Contact(contact, self.labels, resolver._names_map)
            for contact in contacts
        ]


class ContactQuerySet(models.QuerySet):
    def filter_characters(self):
        return self.filter(eve_entity__category=EveEntity.CATEGORY_CHARACTER)

    def filter_corporations(self):
        return self.filter(eve_entity__category=EveEntity.CATEGORY_CORPORATION)

    def filter_alliances(self):
        return self.filter(eve_entity__category=EveEntity.CATEGORY_ALLIANCE)


class AbstractStandingsRequestQuerySet(models.QuerySet):
    def annotate_is_pending(self) -> models.QuerySet:
        return self.annotate(
            is_pending_annotated=Case(
                When(Q(action_date__isnull=True) & Q(is_effective=False), then=True),
                default=Value(False),
                output_field=models.BooleanField(),
            )
        )

    def annotate_is_actioned(self) -> models.QuerySet:
        return self.annotate(
            is_actioned_annotated=Case(
                When(Q(action_date__isnull=False) & Q(is_effective=False), then=True),
                default=Value(False),
                output_field=models.BooleanField(),
            )
        )


class AbstractStandingsRequestManager(models.Manager):
    def filter_characters(self) -> models.QuerySet:
        return self.filter(contact_type_id__in=ContactType.character_ids)

    def filter_corporations(self) -> models.QuerySet:
        return self.filter(contact_type_id__in=ContactType.corporation_ids)

    def get_queryset(self) -> models.QuerySet:
        return AbstractStandingsRequestQuerySet(self.model, using=self._db)

    def process_requests(self) -> None:
        """Process all the Standing requests/revocation objects"""
        from .models import (
            AbstractStandingsRequest,
            StandingRequest,
            StandingRevocation,
        )

        if self.model is AbstractStandingsRequest:
            raise TypeError("Can not be called from abstract objects")

        organization = BaseConfig.standings_source_entity()
        organization_name = organization.name if organization else ""
        for standing_request in self.all():
            contact, dummy = EveEntity.objects.get_or_create_esi(
                id=standing_request.contact_id
            )
            is_currently_effective = standing_request.is_effective
            is_satisfied_standing = standing_request.evaluate_effective_standing()
            if is_satisfied_standing and not is_currently_effective:
                if SR_NOTIFICATIONS_ENABLED:
                    # send notification to user about standing change if enabled
                    if standing_request.is_standing_request:
                        notify(
                            user=standing_request.user,
                            title=_(
                                "%s: Standing with %s now in effect"
                                % (__title__, contact.name)
                            ),
                            message=_(
                                "'%(organization_name)s' now has blue standing with "
                                "your alt %(contact_category)s '%(contact_name)s'. "
                                "Please also update the standing of "
                                "your %(contact_category)s accordingly."
                            )
                            % {
                                "organization_name": organization_name,
                                "contact_category": contact.category,
                                "contact_name": contact.name,
                            },
                        )
                    elif standing_request.is_standing_revocation:
                        if standing_request.user:
                            notify(
                                user=standing_request.user,
                                title="%s: Standing with %s revoked"
                                % (__title__, contact.name),
                                message=_(
                                    "'%(organization_name)s' no longer has "
                                    "standing with your "
                                    "%(contact_category)s '%(contact_name)s'. "
                                    "Please also update the standing of "
                                    "your %(contact_category)s accordingly."
                                )
                                % {
                                    "organization_name": organization_name,
                                    "contact_category": contact.category,
                                    "contact_name": contact.name,
                                },
                            )

                # if this was a revocation the standing requests need to be remove
                # to indicate that this character no longer has standing
                if standing_request.is_standing_revocation:
                    StandingRequest.objects.filter(
                        contact_id=standing_request.contact_id
                    ).delete()
                    StandingRevocation.objects.filter(
                        contact_id=standing_request.contact_id
                    ).delete()

            elif is_satisfied_standing:
                # Just catching all other contact types (corps/alliances)
                # that are set effective
                pass

            elif not is_satisfied_standing and is_currently_effective:
                # Effective standing no longer effective
                logger.info(
                    "Standing for %d is marked as effective but is not "
                    "satisfied in game. Deleting." % standing_request.contact_id
                )
                standing_request.delete(
                    reason=StandingRevocation.Reason.REVOKED_IN_GAME
                )

            else:
                # Check the standing hasn't been set actioned
                # and not updated in game
                actioned_timeout = standing_request.check_actioned_timeout()
                if actioned_timeout is not None and actioned_timeout:
                    logger.info(
                        "Standing request for contact ID %d has timedout "
                        "and will be reset" % standing_request.contact_id
                    )
                    if SR_NOTIFICATIONS_ENABLED:
                        title = _("Standing Request for %s reset" % contact.name)
                        message = _(
                            "The standing request for %(contact_category)s "
                            "'%(contact_name)s' from %(user_name)s "
                            "has been reset as it did not appear in "
                            "game before the timeout period expired."
                            % {
                                "contact_category": contact.category,
                                "contact_name": contact.name,
                                "user_name": standing_request.user.username,
                            },
                        )
                        # Notify standing manager
                        notify(user=actioned_timeout, title=title, message=message)
                        # Notify the user
                        notify(user=standing_request.user, title=title, message=message)

    def has_pending_request(self, contact_id: int) -> bool:
        """Checks if a request is pending for the given contact_id

        contact_id: int contact_id to check the pending request for

        returns True if a request is already pending, False otherwise
        """
        return self.pending_requests().filter(contact_id=contact_id).exists()

    def pending_requests(self) -> models.QuerySet:
        """returns all pending requests for this class"""
        return self.filter(action_date__isnull=True, is_effective=False)


class StandingRequestManager(AbstractStandingsRequestManager):
    def validate_requests(self) -> int:
        """Validate all StandingsRequests and check
        that the user requesting them has permission and has API keys
        associated with the character/corp.

        StandingRevocation are created for invalid standing requests

        returns the number of invalid requests
        """
        from .models import StandingRevocation

        logger.debug("Validating standings requests")
        invalid_count = 0
        for standing_request in self.all():
            logger.debug(
                "Checking request for contact_id %d", standing_request.contact_id
            )
            reason = StandingRevocation.Reason.NONE
            if not standing_request.user.has_perm(self.model.REQUEST_PERMISSION_NAME):
                logger.debug("Request is invalid, user does not have permission")
                reason = StandingRevocation.Reason.LOST_PERMISSION
                is_valid = False

            elif ContactType.is_corporation(
                standing_request.contact_type_id
            ) and not self.model.can_request_corporation_standing(
                standing_request.contact_id, standing_request.user
            ):
                logger.debug("Request is invalid, not all corp API keys recorded.")
                reason = StandingRevocation.Reason.MISSING_CORP_TOKEN
                is_valid = False

            else:
                is_valid = True

            if not is_valid:
                logger.info(
                    "Standing request for contact_id %d no longer valid. "
                    "Creating revocation",
                    standing_request.contact_id,
                )
                StandingRevocation.objects.add_revocation(
                    contact_id=standing_request.contact_id,
                    contact_type=self.model.contact_id_2_type(
                        standing_request.contact_type_id
                    ),
                    user=standing_request.user,
                    reason=reason,
                )
                invalid_count += 1

        return invalid_count

    def create_character_request(
        self, user: User, character: EveCharacter
    ) -> CreateCharacterRequestError:
        """Create new character standings request for user if possible."""
        from .models import ContactSet, RequestLogEntry, StandingRevocation

        try:
            if character.character_ownership.user != user:
                logger.warning(
                    "%s: User %s does not own character, forbidden", character, user
                )
                return CreateCharacterRequestError.USER_IS_NOT_OWNER
        except ObjectDoesNotExist:
            return CreateCharacterRequestError.USER_IS_NOT_OWNER
        try:
            contact_set = ContactSet.objects.latest()
        except ContactSet.DoesNotExist:
            logger.warning("Failed to get a contact set")
            return CreateCharacterRequestError.UNKNOWN_ERROR
        character_id = character.character_id
        if self.has_pending_request(
            character_id
        ) or StandingRevocation.objects.has_pending_request(character_id):
            logger.warning("%s: Character already has a pending request", character)
            return CreateCharacterRequestError.CHARACTER_HAS_REQUEST
        elif not self.model.has_required_scopes_for_request(
            character=character, user=user
        ):
            logger.warning("%s: Character does not have the required scopes", character)
            return CreateCharacterRequestError.CHARACTER_IS_MISSING_SCOPES
        sr = self.get_or_create_2(
            user=user,
            contact_id=character_id,
            contact_type=self.model.ContactType.CHARACTER,
        )
        if contact_set.contact_has_satisfied_standing(character_id):
            sr.mark_actioned(user=None, reason=sr.Reason.STANDING_IN_GAME)
            sr.mark_effective()
            RequestLogEntry.objects.create_from_standing_request(
                sr, RequestLogEntry.Action.CONFIRMED, None
            )
        return CreateCharacterRequestError.NO_ERROR

    def create_corporation_request(self, user, corporation_id) -> bool:
        """Create new corporation standings request for user if possible."""
        from .models import StandingRevocation

        if self.has_pending_request(
            corporation_id
        ) or StandingRevocation.objects.has_pending_request(corporation_id):
            logger.warning(
                "Contact ID %d already has a pending request", corporation_id
            )
            return False
        if not self.model.can_request_corporation_standing(corporation_id, user):
            logger.warning(
                "User %s does not have enough keys for corpID %d, forbidden",
                user,
                corporation_id,
            )
            return False
        self.get_or_create_2(
            user=user,
            contact_id=corporation_id,
            contact_type=self.model.ContactType.CORPORATION,
        )
        return True

    def get_or_create_2(self, user: User, contact_id: int, contact_type: str) -> object:
        """Get or create a new standing request

        Params:
        - user: User the request and contact_id belongs to
        - contact_id: contact_id to request standings on
        - contact_type: type of this contact

        Restuns the created StandingRequest instance
        """
        contact_type_id = self.model.contact_type_2_id(contact_type)
        instance, _ = self.get_or_create(
            contact_id=contact_id,
            contact_type_id=contact_type_id,
            defaults={"user": user},
        )
        return instance


class StandingRevocationManager(AbstractStandingsRequestManager):
    def add_revocation(
        self, contact_id: int, contact_type: str, user: User = None, reason: str = None
    ) -> object:
        """Add a new standings revocation

        Params:
        - contact_id: contact_id to request standings on
        - contact_type_id: contact_type_id from AbstractContact concrete implementation
        - user: user making the request

        Returns the created StandingRevocation instance
        """
        logger.debug(
            "Adding new standings revocation for contact %d type %s",
            contact_id,
            contact_type,
        )
        contact_type_id = self.model.contact_type_2_id(contact_type)
        if self.has_pending_request(contact_id):
            logger.debug(
                "Cannot add revocation for contact %d %s, pending revocation exists",
                contact_id,
                contact_type_id,
            )
            return None
        if not reason:
            reason = self.model.Reason.NONE
        instance = self.create(
            contact_id=contact_id,
            contact_type_id=contact_type_id,
            user=user,
            reason=reason,
        )
        return instance


class CharacterAffiliationManager(models.Manager):
    def update_evecharacter_relations(self) -> None:
        """Update links to eve character in auth if any"""

        eve_character_id_map = {
            obj["character_id"]: obj["id"]
            for obj in EveCharacter.objects.values("id", "character_id")
        }
        with transaction.atomic():
            affiliations = [
                obj for obj in self.filter(character_id__in=eve_character_id_map.keys())
            ]
            for affiliation in affiliations:
                affiliation.eve_character_id = eve_character_id_map[
                    affiliation.character_id
                ]
            self.bulk_update(
                objs=affiliations, fields=["eve_character_id"], batch_size=500
            )

    def update_from_esi(self) -> None:
        """Update all character affiliations we have contacts or requests for."""
        character_ids = self._gather_character_ids()
        if character_ids:
            affiliations = self._fetch_characters_affiliation_from_esi(character_ids)
            if affiliations:
                self._store_affiliations(affiliations)

    def _gather_character_ids(self) -> list:
        from .models import ContactSet, StandingRequest, StandingRevocation

        try:
            contact_set = ContactSet.objects.latest()
        except ContactSet.DoesNotExist:
            logger.warning("Could not find a contact set")
            return []

        character_ids_contacts = set(
            contact_set.contacts.filter_characters()
            .values_list("eve_entity_id", flat=True)
            .distinct()
        )
        character_ids_requests = set(
            StandingRequest.objects.filter_characters()
            .values_list("contact_id", flat=True)
            .distinct()
        )
        character_ids_revocations = set(
            StandingRevocation.objects.filter_characters()
            .values_list("contact_id", flat=True)
            .distinct()
        )
        return list(
            character_ids_contacts | character_ids_requests | character_ids_revocations
        )

    def _fetch_characters_affiliation_from_esi(self, character_ids) -> list:
        chunk_size = 1000
        affiliations = []
        for character_ids_chunk in chunks(character_ids, chunk_size):
            try:
                response = esi.client.Character.post_characters_affiliation(
                    characters=character_ids_chunk
                ).results()
            except HTTPError:
                logger.exception("Could not fetch character affiliations from ESI")
                return []
            else:
                affiliations += response
        return affiliations

    def _store_affiliations(self, affiliations) -> None:
        affiliation_objects = list()
        for affiliation in affiliations:
            character, _ = EveEntity.objects.get_or_create(
                id=affiliation["character_id"]
            )
            corporation, _ = EveEntity.objects.get_or_create(
                id=affiliation["corporation_id"]
            )
            if affiliation.get("alliance_id"):
                alliance, _ = EveEntity.objects.get_or_create(
                    id=affiliation["alliance_id"]
                )
            else:
                alliance = None
            if affiliation.get("faction_id"):
                faction, _ = EveEntity.objects.get_or_create(
                    id=affiliation["faction_id"]
                )
            else:
                faction = None
            affiliation_objects.append(
                self.model(
                    character=character,
                    corporation=corporation,
                    alliance=alliance,
                    faction=faction,
                )
            )
        with transaction.atomic():
            self.all().delete()
            self.bulk_create(affiliation_objects, batch_size=500)
        new_ids = set()
        for obj in affiliation_objects:
            new_ids |= obj.entity_ids()
        EveEntity.objects.bulk_create_esi(new_ids)


class CorporationDetailsManager(models.Manager):
    def corporation_ids_from_contacts(self) -> set:
        from .models import Contact

        contact_corporation_ids = set(
            Contact.objects.filter_corporations().values_list(
                "eve_entity_id", flat=True
            )
        )
        character_affiliation_corporation_ids = set(
            Contact.objects.filter_characters().values_list(
                "eve_entity__character_affiliation__corporation_id", flat=True
            )
        )
        return set(
            filter(
                lambda x: x is not None,
                contact_corporation_ids | character_affiliation_corporation_ids,
            )
        )

    def update_or_create_from_esi(self, id: int) -> Tuple[models.Model, bool]:
        """Updates or create an obj from ESI"""
        logger.info("%s: Fetching corporation from ESI", id)
        data = esi.client.Corporation.get_corporations_corporation_id(
            corporation_id=id
        ).results()
        corporation = EveEntity.objects.get_or_create(id=id)[0]
        alliance = (
            EveEntity.objects.get_or_create(id=data["alliance_id"])[0]
            if data.get("alliance_id")
            else None
        )
        ceo_id = data["ceo_id"] if data["ceo_id"] > 1 else None
        ceo = EveEntity.objects.get_or_create(id=ceo_id)[0] if ceo_id else None
        faction = (
            EveEntity.objects.get_or_create(id=data["faction_id"])[0]
            if data.get("faction_id")
            else None
        )
        EveEntity.objects.bulk_create_esi(
            filter(
                lambda x: x is not None,
                [id, data.get("alliance_id"), ceo_id, data.get("faction_id")],
            )
        )
        return self.update_or_create(
            corporation=corporation,
            defaults={
                "alliance": alliance,
                "ceo": ceo,
                "faction": faction,
                "member_count": data["member_count"],
                "ticker": data["ticker"],
            },
        )


class FrozenQuerySetMixin:
    """Ensures the update method can not be used."""

    def update(self, **kwargs) -> int:
        raise RuntimeError("Update not allowed for this model.")


class RequestLogEntryQuerySet(FrozenQuerySetMixin, models.QuerySet):
    pass


class RequestLogEntryManagerBase(models.Manager):
    def create_from_standing_request(
        self, standing_request, action, action_by
    ) -> Optional[models.Model]:
        from .models import FrozenAlt, FrozenAuthUser

        if standing_request.is_standing_request:
            request_type = self.model.RequestType.REQUEST
        else:
            request_type = self.model.RequestType.REVOCATION
        requested_for, _ = FrozenAlt.objects.get_or_create_from_standing_request(
            standing_request
        )
        if action_by:
            action_by_obj, _ = FrozenAuthUser.objects.get_or_create_from_user(action_by)
        else:
            action_by_obj = None
        requested_by_obj, _ = FrozenAuthUser.objects.get_or_create_from_user(
            standing_request.user
        )
        new_obj = self.create(
            action=action,
            action_by=action_by_obj,
            request_type=request_type,
            requested_at=standing_request.request_date,
            requested_by=requested_by_obj,
            requested_for=requested_for,
            reason=standing_request.reason,
        )
        update_unresolved_eve_entities.delay()
        return new_obj


RequestLogEntryManager = RequestLogEntryManagerBase.from_queryset(
    RequestLogEntryQuerySet
)


class FrozenAuthUserQuerySet(FrozenQuerySetMixin, models.QuerySet):
    pass


class FrozenAuthUserManagerBase(models.Manager):
    def get_or_create_from_user(self, user) -> Tuple[models.Model, bool]:
        main_character = user.profile.main_character
        if main_character:
            character_id = main_character.character_id
            corporation_id = main_character.corporation_id
            alliance_id = main_character.alliance_id
            faction_id = main_character.faction_id
        else:
            character_id = corporation_id = alliance_id = faction_id = None
        try:
            obj = self.get(
                user_id=user.id,
                character_id=character_id,
                corporation_id=corporation_id,
                alliance_id=alliance_id,
                faction_id=faction_id,
            )
            return obj, False
        except self.model.DoesNotExist:
            alliance = (
                EveEntity.objects.get_or_create(id=alliance_id)[0]
                if alliance_id
                else None
            )
            character = (
                EveEntity.objects.get_or_create(id=character_id)[0]
                if character_id
                else None
            )
            corporation = (
                EveEntity.objects.get_or_create(id=corporation_id)[0]
                if corporation_id
                else None
            )
            faction = (
                EveEntity.objects.get_or_create(id=faction_id)[0]
                if faction_id
                else None
            )
            return self.get_or_create(
                user=user,
                character=character,
                corporation=corporation,
                alliance=alliance,
                faction=faction,
                state=user.profile.state,
            )


FrozenAuthUserManager = FrozenAuthUserManagerBase.from_queryset(FrozenAuthUserQuerySet)


class FrozenAltQuerySet(FrozenQuerySetMixin, models.QuerySet):
    pass


class FrozenAltManagerBase(models.Manager):
    def get_or_create_from_standing_request(
        self, standing_request: object
    ) -> Tuple[models.Model, bool]:
        eve_entity, _ = EveEntity.objects.get_or_create(id=standing_request.contact_id)
        if standing_request.is_character:
            category = self.model.Category.CHARACTER
            character = eve_entity
            try:
                alliance = character.character_affiliation.alliance
                corporation = character.character_affiliation.corporation
                faction = character.character_affiliation.faction
            except ObjectDoesNotExist:
                alliance = None
                corporation = None
                faction = None
        elif standing_request.is_corporation:
            category = self.model.Category.CORPORATION
            character = None
            corporation = eve_entity
            try:
                alliance = corporation.corporation_details.alliance
                faction = corporation.corporation_details.faction
            except ObjectDoesNotExist:
                alliance = None
                faction = None
        else:
            raise NotImplementedError()
        return self.get_or_create(
            character=character,
            corporation=corporation,
            alliance=alliance,
            category=category,
            faction=faction,
        )


FrozenAltManager = FrozenAltManagerBase.from_queryset(FrozenAltQuerySet)



# Generated by Django 3.2.9 on 2021-12-01 18:58

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveuniverse", "0005_type_materials_and_sections"),
        ("standingsrequests", "0008_add_revocation_reason"),
    ]

    operations = [
        migrations.AlterField(
            model_name="corporationdetails",
            name="ceo",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
    ]



from django.contrib import messages
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from esi.decorators import token_required
from eveuniverse.models import EveEntity

from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from .. import __title__
from ..app_settings import SR_CORPORATIONS_ENABLED
from ..constants import CreateCharacterRequestError
from ..core import BaseConfig, MainOrganizations
from ..decorators import token_required_by_state
from ..helpers.evecorporation import EveCorporation
from ..models import ContactSet, StandingRequest, StandingRevocation
from ..tasks import update_all, update_associations_api
from .helpers import DEFAULT_ICON_SIZE, add_common_context

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def index_view(request):
    """index page is used as dispatcher"""
    app_count = (
        StandingRequest.objects.pending_requests().count()
        + StandingRevocation.objects.pending_requests().count()
    )
    if app_count > 0 and request.user.has_perm("standingsrequests.affect_standings"):
        return redirect("standingsrequests:manage")
    else:
        return redirect("standingsrequests:create_requests")


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def create_requests(request):
    organization = BaseConfig.standings_source_entity()
    try:
        main_char_id = request.user.profile.main_character.character_id
    except AttributeError:
        main_char_id = None
    context = {
        "corporations_enabled": SR_CORPORATIONS_ENABLED,
        "organization": organization,
        "organization_image_url": organization.icon_url(size=DEFAULT_ICON_SIZE),
        "authinfo": {"main_char_id": main_char_id},
    }
    return render(
        request,
        "standingsrequests/create_requests.html",
        add_common_context(request, context),
    )


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def request_characters(request):
    logger.debug("Start request_characters request")
    try:
        contact_set = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        return render(
            request, "standingsrequests/error.html", add_common_context(request, {})
        )

    eve_characters_qs = EveCharacter.objects.filter(
        character_ownership__user=request.user
    ).select_related("character_ownership__user")
    eve_characters = {obj.character_id: obj for obj in eve_characters_qs}
    characters_with_standing = {
        contact["eve_entity_id"]: contact["standing"]
        for contact in (
            contact_set.contacts.filter(
                eve_entity_id__in=list(eve_characters.keys())
            ).values("eve_entity_id", "standing")
        )
    }
    characters_standings_requests = {
        obj.contact_id: obj
        for obj in (
            StandingRequest.objects.select_related("user")
            .filter(contact_id__in=eve_characters.keys())
            .annotate_is_pending()
            .annotate_is_actioned()
        )
    }
    characters_standing_revocation = {
        obj.contact_id: obj
        for obj in (
            StandingRevocation.objects.filter(
                contact_id__in=eve_characters.keys()
            ).annotate_is_pending()
        )
    }
    characters_data = list()
    for character in eve_characters.values():
        character_id = character.character_id
        standing = characters_with_standing.get(character_id)
        has_pending_request = (
            character_id in characters_standings_requests
            and characters_standings_requests[character_id].is_pending_annotated
        )
        has_pending_revocation = (
            character_id in characters_standing_revocation
            and characters_standing_revocation[character_id].is_pending_annotated
        )
        has_actioned_request = (
            character_id in characters_standings_requests
            and characters_standings_requests[character_id].is_actioned_annotated
        )
        has_standing = (
            character_id in characters_standings_requests
            and characters_standings_requests[character_id].is_effective
            and characters_standings_requests[character_id].user == request.user
        )
        characters_data.append(
            {
                "character": character,
                "standing": standing,
                "pendingRequest": has_pending_request,
                "pendingRevocation": has_pending_revocation,
                "requestActioned": has_actioned_request,
                "inOrganisation": MainOrganizations.is_character_a_member(character),
                "hasRequiredScopes": StandingRequest.has_required_scopes_for_request(
                    character, user=request.user, quick_check=True
                ),
                "hasStanding": has_standing,
            }
        )

    context = {"characters": characters_data}
    return render(
        request,
        "standingsrequests/partials/_request_characters.html",
        add_common_context(request, context),
    )


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def request_corporations(request):
    logger.debug("Start request_characters request")
    try:
        contact_set = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        return render(
            request, "standingsrequests/error.html", add_common_context(request, {})
        )

    eve_characters_qs = EveCharacter.objects.filter(
        character_ownership__user=request.user
    ).select_related("character_ownership__user")
    corporation_ids = set(
        eve_characters_qs.exclude(corporation_id__in=MainOrganizations.corporation_ids)
        .exclude(alliance_id__in=MainOrganizations.alliance_ids)
        .values_list("corporation_id", flat=True)
    )
    corporations_standing_requests = {
        obj.contact_id: obj
        for obj in (
            StandingRequest.objects.select_related("user")
            .filter(contact_id__in=corporation_ids)
            .annotate_is_pending()
            .annotate_is_actioned()
        )
    }
    corporations_revocation_requests = {
        obj.contact_id: obj
        for obj in (
            StandingRevocation.objects.filter(contact_id__in=corporation_ids)
            .annotate_is_pending()
            .annotate_is_actioned()
        )
    }
    corporation_contacts = {
        obj.eve_entity_id: obj
        for obj in (contact_set.contacts.filter(eve_entity_id__in=corporation_ids))
    }
    corporations_data = list()
    for corporation in EveCorporation.get_many_by_id(corporation_ids):
        if corporation and not corporation.is_npc:
            corporation_id = corporation.corporation_id
            try:
                standing = corporation_contacts[corporation_id].standing
            except KeyError:
                standing = None
            has_pending_request = (
                corporation_id in corporations_standing_requests
                and corporations_standing_requests[corporation_id].is_pending_annotated
            )
            has_pending_revocation = (
                corporation_id in corporations_revocation_requests
                and corporations_revocation_requests[
                    corporation_id
                ].is_pending_annotated
            )
            has_actioned_request = (
                corporation_id in corporations_standing_requests
                and corporations_standing_requests[corporation_id].is_actioned_annotated
            )
            has_standing = (
                corporation_id in corporations_standing_requests
                and corporations_standing_requests[corporation_id].is_effective
                and corporations_standing_requests[corporation_id].user == request.user
            )
            corporations_data.append(
                {
                    "token_count": corporation.member_tokens_count_for_user(
                        request.user, quick_check=True
                    ),
                    "corp": corporation,
                    "standing": standing,
                    "pendingRequest": has_pending_request,
                    "pendingRevocation": has_pending_revocation,
                    "requestActioned": has_actioned_request,
                    "hasStanding": has_standing,
                }
            )

    corporations_data.sort(key=lambda x: x["corp"].corporation_name)
    context = {"corps": corporations_data}
    return render(
        request,
        "standingsrequests/partials/_request_corporations.html",
        add_common_context(request, context),
    )


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def request_character_standing(request, character_id: int):
    """For a user to request standings for their own characters"""
    logger.debug(
        "Standings request from user %s for characterID %s",
        str(request.user),
        character_id,
    )
    character = get_object_or_404(
        EveCharacter.objects.select_related("character_ownership__user"),
        character_id=character_id,
    )
    error = StandingRequest.objects.create_character_request(request.user, character)
    if error is not CreateCharacterRequestError.NO_ERROR:
        if error is CreateCharacterRequestError.CHARACTER_IS_MISSING_SCOPES:
            messages.error(
                request,
                "You character %s is missing scopes."
                % EveEntity.objects.resolve_name(character_id),
            )
        elif error is CreateCharacterRequestError.USER_IS_NOT_OWNER:
            messages.error(
                request,
                "You are not the owner of character %s."
                % EveEntity.objects.resolve_name(character_id),
            )
        else:
            messages.error(
                request,
                "An unexpected error occurred when trying to process "
                "your standing request for %s. Please try again."
                % EveEntity.objects.resolve_name(character_id),
            )
    else:
        update_associations_api.delay()
    return redirect("standingsrequests:create_requests")


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def remove_character_standing(request, character_id: int):
    """
    Handles both removing requests and removing existing standings
    """
    logger.debug(
        "remove_character_standing called by %s for character %s",
        str(request.user),
        character_id,
    )
    req = get_object_or_404(StandingRequest, user=request.user, contact_id=character_id)
    success = req.remove()
    if not success:
        messages.warning(
            request,
            "An unexpected error occurred when trying to process "
            "your request to revoke standing for %s. Please try again."
            % EveEntity.objects.resolve_name(character_id),
        )
    return redirect("standingsrequests:create_requests")


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def request_corp_standing(request, corporation_id):
    """
    For a user to request standings for their own corp
    """
    corporation_id = int(corporation_id)
    logger.debug(
        "Standings request from user %s for corpID %d", request.user, corporation_id
    )
    if not StandingRequest.objects.create_corporation_request(
        request.user, corporation_id
    ):
        messages.warning(
            request,
            "An unexpected error occurred when trying to process "
            "your standing request for %s. Please try again."
            % EveEntity.objects.resolve_name(corporation_id),
        )
    else:
        update_associations_api.delay()
    return redirect("standingsrequests:create_requests")


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def remove_corp_standing(request, corporation_id: int):
    """
    Handles both removing corp requests and removing existing standings
    """
    logger.debug("remove_corp_standing called by %s", request.user)
    try:
        req = StandingRequest.objects.filter(user=request.user).get(
            contact_id=corporation_id
        )
    except StandingRequest.DoesNotExist:
        success = False
    else:
        success = req.remove()
    if not success:
        messages.warning(
            request,
            "An unexpected error occurred when trying to process "
            "your request to revoke standing for %s. Please try again."
            % EveEntity.objects.resolve_name(corporation_id),
        )
    return redirect("standingsrequests:create_requests")


@login_required
@permission_required("standingsrequests.affect_standings")
@token_required(new=False, scopes=ContactSet.required_esi_scope())
def view_auth_page(request, token):
    source_entity = BaseConfig.standings_source_entity()
    char_name = EveEntity.objects.resolve_name(BaseConfig.owner_character_id)
    if not source_entity:
        messages.error(
            request,
            format_html(
                _(
                    "The configured character <strong>%s</strong> does not belong "
                    "to an alliance and can therefore not be used "
                    "to setup alliance standings. "
                    "Please configure a character that has an alliance."
                )
                % char_name,
            ),
        )
    elif token.character_id == BaseConfig.owner_character_id:
        update_all.delay(user_pk=request.user.pk)
        messages.success(
            request,
            format_html(
                _(
                    "Token for character <strong>%s</strong> has been setup "
                    "successfully and the app has started pulling standings "
                    "from <strong>%s</strong>."
                )
                % (char_name, source_entity.name),
            ),
        )
    else:
        messages.error(
            request,
            _(
                "Failed to setup token for configured character "
                "%(char_name)s (id:%(standings_api_char_id)s). "
                "Instead got token for different character: "
                "%(token_char_name)s (id:%(token_char_id)s)"
            )
            % {
                "char_name": char_name,
                "standings_api_char_id": BaseConfig.owner_character_id,
                "token_char_name": EveEntity.objects.resolve_name(token.character_id),
                "token_char_id": token.character_id,
            },
        )
    return redirect("standingsrequests:index")


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
@token_required_by_state(new=False)
def view_requester_add_scopes(request, token):
    messages.success(
        request,
        _("Successfully added token with required scopes for %(char_name)s")
        % {"char_name": EveEntity.objects.resolve_name(token.character_id)},
    )
    return redirect("standingsrequests:create_requests")


@login_required
@staff_member_required
def admin_changeset_update_now(request):
    update_all.delay(user_pk=request.user.pk)
    messages.info(
        request,
        _(
            "Started updating contacts and affiliations. "
            "You will receive a notification when completed."
        ),
    )
    return redirect("admin:standingsrequests_contactset_changelist")



from concurrent.futures import ThreadPoolExecutor

from bravado.exception import HTTPError

from django.contrib.auth.models import User
from django.core.cache import cache
from eveuniverse.models import EveEntity

from allianceauth.eveonline.evelinks import eveimageserver
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from .. import __title__
from ..providers import esi

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

MAX_WORKERS = 10


class EveCorporation:
    CACHE_PREFIX = "STANDINGS_REQUESTS_EVECORPORATION_"
    CACHE_TIME = 60 * 60  # 60 minutes

    def __init__(self, **kwargs):
        self.corporation_id = int(kwargs.get("corporation_id"))
        self.corporation_name = kwargs.get("corporation_name")
        self.ticker = kwargs.get("ticker")
        self.member_count = kwargs.get("member_count")
        self.ceo_id = kwargs.get("ceo_id")
        self.alliance_id = kwargs.get("alliance_id")
        self.alliance_name = kwargs.get("alliance_name")

    def __str__(self):
        return self.corporation_name

    def __eq__(self, o: object) -> bool:
        return (
            isinstance(o, type(self))
            and self.corporation_id == o.corporation_id
            and self.corporation_name == o.corporation_name
            and self.ticker == o.ticker
            and self.member_count == o.member_count
            and self.ceo_id == o.ceo_id
            and self.alliance_id == o.alliance_id
            and self.alliance_name == o.alliance_name
        )

    @property
    def is_npc(self) -> bool:
        """returns true if this corporation is an NPC, else false"""
        return self.corporation_is_npc(self.corporation_id)

    @staticmethod
    def corporation_is_npc(corporation_id: int) -> bool:
        """returns true if this corporation is an NPC, else false"""
        return 1000000 <= corporation_id <= 2000000

    def logo_url(self, size: int = eveimageserver._DEFAULT_IMAGE_SIZE) -> str:
        return eveimageserver.corporation_logo_url(self.corporation_id, size)

    def member_tokens_count_for_user(
        self, user: User, quick_check: bool = False
    ) -> int:
        """returns the number of character tokens the given user owns
        for this corporation

        Params:
        - user: user owning the characters
        - quick: if True will not check if tokens are valid to save time
        """
        from ..models import StandingRequest

        return sum(
            [
                1
                for character in EveCharacter.objects.filter(
                    character_ownership__user=user
                )
                .select_related("character_ownership__user__profile__state")
                .filter(corporation_id=self.corporation_id)
                if StandingRequest.has_required_scopes_for_request(
                    character=character, user=user, quick_check=quick_check
                )
            ]
        )

    def user_has_all_member_tokens(self, user: User, quick_check: bool = False) -> bool:
        """returns True if given user owns same amount of token than there are
        member characters in this corporation, else False

        Params:
        - user: user owning the characters
        - quick: if True will not check if tokens are valid to save time
        """
        return (
            self.member_count is not None
            and self.member_tokens_count_for_user(user=user, quick_check=quick_check)
            >= self.member_count
        )

    @classmethod
    def get_by_id(cls, corporation_id: int, ignore_cache: bool = False) -> object:
        """Get a corporation from the cache or ESI if not cached
        Corps are cached for 3 hours

        Params
        - corporation_id: int corporation ID to get
        - ignore_cache: when true will always get fresh from API

        Returns corporation object or None
        """
        logger.debug("Getting corporation by id %d", corporation_id)
        my_cache_key = cls._get_cache_key(corporation_id)
        corporation = cache.get(my_cache_key)
        if corporation is None or ignore_cache:
            logger.debug("Corp not in cache or ignoring cache, fetching")
            corporation = cls.fetch_corporation_from_api(corporation_id)
            if corporation is not None:
                cache.set(my_cache_key, corporation, cls.CACHE_TIME)
        else:
            logger.debug("Retreving corporation %s from cache", corporation_id)
        return corporation

    @classmethod
    def _get_cache_key(cls, corporation_id: int) -> str:
        return cls.CACHE_PREFIX + str(corporation_id)

    @classmethod
    def fetch_corporation_from_api(cls, corporation_id):
        logger.debug(
            "Attempting to fetch corporation from ESI with id %s", corporation_id
        )
        try:
            info = esi.client.Corporation.get_corporations_corporation_id(
                corporation_id=corporation_id
            ).results()
        except HTTPError:
            logger.exception(
                "Failed to fetch corporation from ESI with id %i", corporation_id
            )
            return None

        else:
            args = {
                "corporation_id": corporation_id,
                "corporation_name": info["name"],
                "ticker": info["ticker"],
                "member_count": info["member_count"],
                "ceo_id": info["ceo_id"],
            }
            if "alliance_id" in info and info["alliance_id"]:
                args["alliance_id"] = info["alliance_id"]
                args["alliance_name"] = EveEntity.objects.resolve_name(
                    info["alliance_id"]
                )

            return cls(**args)

    @classmethod
    def thread_fetch_corporation(cls, corporation_id: int) -> object:
        """Gets one corporation by ID and returns it - used for threads"""
        return EveCorporation.get_by_id(corporation_id)

    @classmethod
    def get_many_by_id(cls, corporation_ids: list) -> list:
        """Returns multiple corporations by ID

        Fetches requested corporations from cache or API as needed.
        Uses threads to fetch them in parallel.
        """
        corporation_ids = set(corporation_ids)
        if not corporation_ids:
            return []

        # make sure client is loaded before starting threads
        esi.client
        logger.info(
            "Starting to fetch the %d corporations from ESI with up to %d workers",
            len(corporation_ids),
            MAX_WORKERS,
        )
        with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
            futures = [
                executor.submit(cls.thread_fetch_corporation, corporation_id)
                for corporation_id in corporation_ids
            ]
            logger.info("Waiting for all threads fetching corporations to complete...")

        logger.info("Completed fetching %d corporations from ESI", len(corporation_ids))
        return [f.result() for f in futures]



from .. import __title__
from ..core import BaseConfig
from ..models import StandingRequest, StandingRevocation

DEFAULT_ICON_SIZE = 32


def add_common_context(request, context: dict) -> dict:
    """adds the common context used by all view"""
    new_context = {
        **{
            "app_title": __title__,
            "operation_mode": str(BaseConfig.operation_mode),
            "pending_total_count": (
                StandingRequest.objects.pending_requests().count()
                + StandingRevocation.objects.pending_requests().count()
            ),
        },
        **context,
    }
    return new_context



from datetime import timedelta

from celery import chain, shared_task

from django.contrib.auth.models import User
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from eveuniverse.tasks import update_unresolved_eve_entities

from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import SR_STANDINGS_STALE_HOURS, SR_SYNC_BLUE_ALTS_ENABLED
from .core import BaseConfig
from .models import (
    CharacterAffiliation,
    Contact,
    ContactLabel,
    ContactSet,
    CorporationDetails,
    StandingRequest,
    StandingRevocation,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@shared_task(name="standings_requests.update_all")
def update_all(user_pk: int = None):
    """Updates standings and affiliations"""
    my_chain = chain(
        [
            standings_update.si(),
            update_associations_api.si(),
            report_result_to_user.si(user_pk),
        ]
    )
    my_chain.delay()


@shared_task(name="standings_requests.report_result_to_user")
def report_result_to_user(user_pk: int = None):
    if user_pk:
        try:
            user = User.objects.get(pk=user_pk)
        except User.DoesNotExist:
            logger.warning("Can not find a user with pk %d", user_pk)
            return
        else:
            source_entity = BaseConfig.standings_source_entity()
            notify(
                user,
                _("%s: Standings loaded" % __title__),
                _(
                    "Standings have been successfully loaded for %s"
                    % source_entity.name
                ),
                level="success",
            )


@shared_task(name="standings_requests.standings_update")
def standings_update():
    """Updates standings from ESI"""
    logger.info("Standings API update started")
    contact_set = ContactSet.objects.create_new_from_api()
    if not contact_set:
        logger.warn(
            "Standings API update returned None (API error probably),"
            "aborting standings update"
        )
    else:
        if SR_SYNC_BLUE_ALTS_ENABLED:
            contact_set.generate_standing_requests_for_blue_alts()
        StandingRequest.objects.process_requests()
        StandingRevocation.objects.process_requests()


@shared_task(name="standings_requests.validate_requests")
def validate_requests():
    logger.info("Validating standings request running")
    count = StandingRequest.objects.validate_requests()
    logger.info("Dealt with %d invalid standings requests", count)


@shared_task(name="standings_requests.update_associations_auth")
def update_associations_auth():
    ...


@shared_task(name="standings_requests.update_associations_api")
def update_associations_api():
    """Update character affiliations from ESI and relations to Eve Characters"""
    chain(
        [
            _update_character_affiliations_from_esi.si(),
            _update_character_affiliations_to_auth.si(),
            update_all_corporation_details.si(),
            update_unresolved_eve_entities.si(),
        ]
    ).delay()


@shared_task
def _update_character_affiliations_from_esi():
    logger.info("Running character affiliations updating from ESI...")
    CharacterAffiliation.objects.update_from_esi()
    logger.info("Finished character affiliations from ESI.")


@shared_task
def _update_character_affiliations_to_auth():
    logger.info("Updating character affiliations relations to Auth...")
    CharacterAffiliation.objects.update_evecharacter_relations()
    logger.info("Finished updating character affiliations to Auth.")


@shared_task(name="standings_requests.purge_stale_data")
def purge_stale_data():
    """Delete all the data which is beyond its useful life.
    There is no harm in disabling this if you wish to keep everything.
    """
    my_chain = chain([purge_stale_standings_data.si()])
    my_chain.delay()


@shared_task
def purge_stale_standings_data():
    """Deletes all stale (=older than threshold hours) contact sets
    except the last remaining contact set
    """
    logger.info("Purging stale standings data")
    cutoff_date = now() - timedelta(hours=SR_STANDINGS_STALE_HOURS)
    try:
        latest_standings = ContactSet.objects.latest()
        stale_contacts_qs = ContactSet.objects.filter(date__lt=cutoff_date).exclude(
            id=latest_standings.id
        )
        if stale_contacts_qs.exists():
            logger.debug("Deleting old ContactSets")
            # we can't just do standigs.delete()
            # because with lots of them it uses lots of memory
            # lets go over them one by one and delete
            for contact_set in stale_contacts_qs:
                Contact.objects.filter(contact_set=contact_set).delete()
                Contact.objects.filter(contact_set=contact_set).delete()
                Contact.objects.filter(contact_set=contact_set).delete()
                ContactLabel.objects.filter(contact_set=contact_set).delete()

            stale_contacts_qs.delete()
        else:
            logger.debug("No ContactSets to delete")

    except ContactSet.DoesNotExist:
        logger.warn("No ContactSets available, nothing to delete")


@shared_task
def update_all_corporation_details():
    existing_corporation_ids = (
        CorporationDetails.objects.corporation_ids_from_contacts()
    )
    CorporationDetails.objects.exclude(
        corporation_id__in=existing_corporation_ids
    ).delete()
    if existing_corporation_ids:
        logger.info(
            "Updating corporation details for %d corporations.",
            len(existing_corporation_ids),
        )
        for corporation_id in existing_corporation_ids:
            update_corporation_detail.delay(corporation_id)
    else:
        logger.info("No corporations to update.")


@shared_task
def update_corporation_detail(corporation_id: int):
    CorporationDetails.objects.update_or_create_from_esi(corporation_id)



# Generated by Django 3.1.10 on 2021-05-12 19:51

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveuniverse", "0005_type_materials_and_sections"),
        ("eveonline", "0014_auto_20210105_1413"),
        ("standingsrequests", "0006_add_user_to_revocations"),
    ]

    operations = [
        migrations.CreateModel(
            name="CharacterAffiliation",
            fields=[
                (
                    "character",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="character_affiliation",
                        serialize=False,
                        to="eveuniverse.eveentity",
                    ),
                ),
                ("updated", models.DateTimeField(auto_now_add=True)),
                (
                    "alliance",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "eve_character",
                    models.ForeignKey(
                        default=None,
                        help_text="Related auth character (if any)",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveonline.evecharacter",
                    ),
                ),
                (
                    "faction",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Contact",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("standing", models.FloatField(db_index=True)),
                ("is_watched", models.BooleanField(default=False)),
                (
                    "contact_set",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="contacts",
                        to="standingsrequests.contactset",
                    ),
                ),
                (
                    "eve_entity",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="standingrequests_contact",
                        to="eveuniverse.eveentity",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="CorporationDetails",
            fields=[
                (
                    "corporation",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="corporation_details",
                        serialize=False,
                        to="eveuniverse.eveentity",
                    ),
                ),
                ("member_count", models.PositiveIntegerField()),
                ("ticker", models.CharField(max_length=255)),
                (
                    "alliance",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "ceo",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "faction",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
            ],
        ),
        migrations.RemoveField(
            model_name="alliancecontact",
            name="contact_set",
        ),
        migrations.RemoveField(
            model_name="alliancecontact",
            name="labels",
        ),
        migrations.DeleteModel(
            name="CharacterAssociation",
        ),
        migrations.RemoveField(
            model_name="charactercontact",
            name="contact_set",
        ),
        migrations.RemoveField(
            model_name="charactercontact",
            name="labels",
        ),
        migrations.RemoveField(
            model_name="corporationcontact",
            name="contact_set",
        ),
        migrations.RemoveField(
            model_name="corporationcontact",
            name="labels",
        ),
        migrations.DeleteModel(
            name="EveEntity",
        ),
        migrations.AlterField(
            model_name="contactlabel",
            name="contact_set",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="labels",
                to="standingsrequests.contactset",
            ),
        ),
        migrations.DeleteModel(
            name="AllianceContact",
        ),
        migrations.DeleteModel(
            name="CharacterContact",
        ),
        migrations.DeleteModel(
            name="CorporationContact",
        ),
        migrations.AddField(
            model_name="contact",
            name="labels",
            field=models.ManyToManyField(
                related_name="contacts", to="standingsrequests.ContactLabel"
            ),
        ),
    ]



from django.contrib.auth.decorators import login_required, permission_required
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.http import HttpResponse, HttpResponseNotFound, JsonResponse
from django.shortcuts import get_object_or_404, render
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import cache_page
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag
from app_utils.views import HttpResponseNoContent

from .. import __title__
from ..app_settings import SR_NOTIFICATIONS_ENABLED, SR_PAGE_CACHE_SECONDS
from ..core import BaseConfig, ContactType
from ..helpers.evecharacter import EveCharacterHelper
from ..helpers.evecorporation import EveCorporation
from ..helpers.writers import UnicodeWriter
from ..models import ContactSet, RequestLogEntry, StandingRequest, StandingRevocation
from .helpers import DEFAULT_ICON_SIZE, add_common_context

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


###########################
# Views character and groups #
###########################
@login_required
@permission_required("standingsrequests.view")
def view_pilots_standings(request):
    logger.debug("view_pilot_standings called by %s", request.user)
    try:
        contact_set = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contact_set = None
    finally:
        organization = BaseConfig.standings_source_entity()
        last_update = contact_set.date if contact_set else None
        pilots_count = contact_set.contacts.count() if contact_set else None

    context = {
        "lastUpdate": last_update,
        "organization": organization,
        "pilots_count": pilots_count,
    }
    return render(
        request,
        "standingsrequests/view_pilots.html",
        add_common_context(request, context),
    )


@cache_page(SR_PAGE_CACHE_SECONDS)
@login_required
@permission_required("standingsrequests.view")
def view_pilots_standings_json(request):
    try:
        contacts = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contacts = ContactSet()
    character_contacts_qs = (
        contacts.contacts.filter_characters()
        .select_related(
            "eve_entity",
            "eve_entity__character_affiliation",
            "eve_entity__character_affiliation__corporation",
            "eve_entity__character_affiliation__alliance",
            "eve_entity__character_affiliation__faction",
            "eve_entity__character_affiliation__eve_character",
            "eve_entity__character_affiliation__eve_character__character_ownership__user",
            "eve_entity__character_affiliation__eve_character__character_ownership__user__profile__main_character",
            "eve_entity__character_affiliation__eve_character__character_ownership__user__profile__state",
        )
        .prefetch_related("labels")
        .order_by("eve_entity__name")
    )
    characters_data = list()
    for contact in character_contacts_qs:
        try:
            character = contact.eve_entity.character_affiliation.eve_character
            user = character.character_ownership.user
        except (AttributeError, ObjectDoesNotExist):
            main = None
            state = ""
            main_character_name = None
            main_character_ticker = None
            main_character_icon_url = None
        else:
            main = user.profile.main_character
            state = user.profile.state.name if user.profile.state else ""
            main_character_name = main.character_name
            main_character_ticker = main.corporation_ticker
            main_character_icon_url = main.portrait_url(DEFAULT_ICON_SIZE)
        try:
            assoc = contact.eve_entity.character_affiliation
        except (AttributeError, ObjectDoesNotExist):
            corporation_id = None
            corporation_name = "?"
            alliance_id = None
            alliance_name = "?"
            faction_id = None
            faction_name = "?"
        else:
            corporation_id = assoc.corporation.id
            corporation_name = assoc.corporation.name
            alliance_id = assoc.alliance.id if assoc.alliance else None
            alliance_name = assoc.alliance.name if assoc.alliance else ""
            faction_id = assoc.faction.id if assoc.faction else None
            faction_name = assoc.faction.name if assoc.faction else None

        labels = [label.name for label in contact.labels.all()]
        characters_data.append(
            {
                "character_id": contact.eve_entity_id,
                "character_name": contact.eve_entity.name,
                "character_icon_url": contact.eve_entity.icon_url(DEFAULT_ICON_SIZE),
                "corporation_id": corporation_id,
                "corporation_name": corporation_name,
                "alliance_id": alliance_id,
                "alliance_name": alliance_name,
                "faction_id": faction_id,
                "faction_name": faction_name,
                "state": state,
                "main_character_name": main_character_name,
                "main_character_ticker": main_character_ticker,
                "main_character_icon_url": main_character_icon_url,
                "standing": contact.standing,
                "labels": labels,
            }
        )
    return JsonResponse(characters_data, safe=False)


@login_required
@permission_required("standingsrequests.download")
def download_pilot_standings(request):
    logger.info("download_pilot_standings called by %s", request.user)
    response = HttpResponse(content_type="text/csv")
    response["Content-Disposition"] = 'attachment; filename="standings.csv"'
    writer = UnicodeWriter(response)
    try:
        contacts = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contacts = ContactSet()
    writer.writerow(
        [
            "character_id",
            "character_name",
            "corporation_id",
            "corporation_name",
            "corporation_ticker",
            "alliance_id",
            "alliance_name",
            "has_scopes",
            "state",
            "main_character_name",
            "main_character_ticker",
            "standing",
            "labels",
        ]
    )

    # lets request make sure all info is there in bulk
    character_contacts = contacts.contacts.all().order_by("eve_entity__name")
    EveEntity.objects.bulk_resolve_names([p.contact_id for p in character_contacts])

    for pilot_standing in character_contacts:
        try:
            char = EveCharacter.objects.get(character_id=pilot_standing.contact_id)
        except EveCharacter.DoesNotExist:
            char = None
        main = ""
        state = ""
        try:
            ownership = CharacterOwnership.objects.get(character=char)
        except CharacterOwnership.DoesNotExist:
            main_character_name = ""
            main = None
        else:
            state = ownership.user.profile.state.name
            main = ownership.user.profile.main_character
            if main is None:
                main_character_name = ""
            else:
                main_character_name = main.character_name
        pilot = [
            pilot_standing.eve_entity_id,
            pilot_standing.eve_entity.name,
            char.corporation_id if char else "",
            char.corporation_name if char else "",
            char.corporation_ticker if char else "",
            char.alliance_id if char else "",
            char.alliance_name if char else "",
            StandingRequest.has_required_scopes_for_request(char),
            state,
            main_character_name,
            main.corporation_ticker if main else "",
            pilot_standing.standing,
            ", ".join([label.name for label in pilot_standing.labels.all()]),
        ]
        writer.writerow([str(v) if v is not None else "" for v in pilot])
    return response


@login_required
@permission_required("standingsrequests.view")
def view_groups_standings(request):
    logger.debug("view_group_standings called by %s", request.user)
    try:
        contact_set = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contact_set = None
    finally:
        organization = BaseConfig.standings_source_entity()
        last_update = contact_set.date if contact_set else None

    if contact_set:
        groups_count = (
            contact_set.contacts.filter_corporations()
            | contact_set.contacts.filter_alliances()
        ).count()

    else:
        groups_count = None
    context = {
        "lastUpdate": last_update,
        "organization": organization,
        "groups_count": groups_count,
    }
    return render(
        request,
        "standingsrequests/view_groups.html",
        add_common_context(request, context),
    )


@cache_page(SR_PAGE_CACHE_SECONDS)
@login_required
@permission_required("standingsrequests.view")
def view_groups_standings_json(request):
    try:
        contacts = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contacts = ContactSet()
    corporations_qs = (
        contacts.contacts.filter_corporations()
        .select_related(
            "eve_entity",
            "eve_entity__corporation_details",
            "eve_entity__corporation_details__alliance",
            "eve_entity__corporation_details__faction",
        )
        .prefetch_related("labels")
        .order_by("eve_entity__name")
    )
    corporations_data = list()
    standings_requests = {
        obj.contact_id: obj
        for obj in (
            StandingRequest.objects.filter(
                contact_type_id=ContactType.corporation_id
            ).filter(
                contact_id__in=list(
                    corporations_qs.values_list("eve_entity_id", flat=True)
                )
            )
        )
    }
    for contact in corporations_qs:
        try:
            corporation_details = contact.eve_entity.corporation_details
        except (ObjectDoesNotExist, AttributeError):
            alliance_id = None
            alliance_name = "?"
            faction_id = None
            faction_name = "?"
        else:
            alliance = corporation_details.alliance
            if alliance:
                alliance_id = alliance.id
                alliance_name = alliance.name
            else:
                alliance_id = None
                alliance_name = ""
            faction = corporation_details.faction
            if faction:
                faction_id = faction.id
                faction_name = faction.name
            else:
                faction_id = None
                faction_name = ""
        try:
            standing_request = standings_requests[contact.eve_entity_id]
            user = standing_request.user
            main = user.profile.main_character
        except (KeyError, AttributeError, ObjectDoesNotExist):
            main_character_name = ""
            main_character_ticker = ""
            main_character_icon_url = ""
            state_name = ""
        else:
            main_character_name = main.character_name if main else ""
            main_character_ticker = main.corporation_ticker if main else ""
            main_character_icon_url = (
                main.portrait_url(DEFAULT_ICON_SIZE) if main else ""
            )
            state_name = user.profile.state.name

        labels = [label.name for label in contact.labels.all()]
        corporations_data.append(
            {
                "corporation_id": contact.eve_entity_id,
                "corporation_name": contact.eve_entity.name,
                "corporation_icon_url": contact.eve_entity.icon_url(DEFAULT_ICON_SIZE),
                "alliance_id": alliance_id,
                "alliance_name": alliance_name,
                "faction_id": faction_id,
                "faction_name": faction_name,
                "standing": contact.standing,
                "labels": labels,
                "state": state_name,
                "main_character_name": main_character_name,
                "main_character_ticker": main_character_ticker,
                "main_character_icon_url": main_character_icon_url,
            }
        )
    alliances_data = list()
    for contact in (
        contacts.contacts.filter_alliances()
        .select_related("eve_entity")
        .prefetch_related("labels")
        .order_by("eve_entity__name")
    ):
        alliances_data.append(
            {
                "alliance_id": contact.eve_entity_id,
                "alliance_name": contact.eve_entity.name,
                "alliance_icon_url": contact.eve_entity.icon_url(DEFAULT_ICON_SIZE),
                "standing": contact.standing,
                "labels": [label.name for label in contact.labels.all()],
            }
        )
    my_groups_data = {"corps": corporations_data, "alliances": alliances_data}
    return JsonResponse(my_groups_data, safe=False)


###################
# Manage requests #
###################


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_standings(request):
    logger.debug("manage_standings called by %s", request.user)
    context = {
        "organization": BaseConfig.standings_source_entity(),
        "requests_count": StandingRequest.objects.pending_requests().count(),
        "revocations_count": StandingRevocation.objects.pending_requests().count(),
    }
    return render(
        request, "standingsrequests/manage.html", add_common_context(request, context)
    )


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_get_requests_json(request):
    logger.debug("manage_get_requests_json called by %s", request.user)
    requests_qs = StandingRequest.objects.pending_requests()
    requests_data = _compose_standing_requests_data(requests_qs)
    return JsonResponse(requests_data, safe=False)


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_get_revocations_json(request):
    logger.debug("manage_get_revocations_json called by %s", request.user)
    revocations_qs = StandingRevocation.objects.pending_requests()
    requests_data = _compose_standing_requests_data(revocations_qs)
    return JsonResponse(requests_data, safe=False)


def _compose_standing_requests_data(
    requests_qs: models.QuerySet, quick_check: bool = False
) -> list:
    """composes list of standings requests or revocations based on queryset
    and returns it
    """
    requests_qs = requests_qs.select_related(
        "user", "user__profile__state", "user__profile__main_character"
    )
    # preload data in bulk
    eve_characters = {
        character.character_id: character
        for character in EveCharacter.objects.filter(
            character_id__in=(
                requests_qs.exclude(
                    contact_type_id=ContactType.corporation_id
                ).values_list("contact_id", flat=True)
            )
        )
    }
    # TODO: remove EveCorporation usage
    eve_corporations = {
        corporation.corporation_id: corporation
        for corporation in EveCorporation.get_many_by_id(
            requests_qs.filter(contact_type_id=ContactType.corporation_id).values_list(
                "contact_id", flat=True
            )
        )
    }
    try:
        contact_set = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contacts = dict()
    else:
        all_contact_ids = set(eve_characters.keys()) | set(eve_corporations.keys())
        contacts = {
            obj.eve_entity_id: obj
            for obj in contact_set.contacts.prefetch_related("labels").filter(
                eve_entity_id__in=all_contact_ids
            )
        }
    requests_data = list()
    for req in requests_qs:
        main_character_name = ""
        main_character_ticker = ""
        main_character_icon_url = ""
        if req.user:
            state_name = req.user.profile.state.name
            main = req.user.profile.main_character
            if main:
                main_character_name = main.character_name
                main_character_ticker = main.corporation_ticker
                main_character_icon_url = main.portrait_url(DEFAULT_ICON_SIZE)
        else:
            state_name = "(no user)"

        if req.is_character:
            if req.contact_id in eve_characters:
                character = eve_characters[req.contact_id]
            else:
                # TODO: remove EveCharacterHelper usage
                character = EveCharacterHelper(req.contact_id)

            contact_name = character.character_name
            contact_icon_url = character.portrait_url(DEFAULT_ICON_SIZE)
            corporation_id = character.corporation_id
            corporation_name = (
                character.corporation_name if character.corporation_name else ""
            )
            corporation_ticker = (
                character.corporation_ticker if character.corporation_ticker else ""
            )
            alliance_id = character.alliance_id
            alliance_name = character.alliance_name if character.alliance_name else ""
            has_scopes = StandingRequest.has_required_scopes_for_request(
                character=character, user=req.user, quick_check=quick_check
            )

        elif req.is_corporation and req.contact_id in eve_corporations:
            corporation = eve_corporations[req.contact_id]
            contact_icon_url = corporation.logo_url(DEFAULT_ICON_SIZE)
            contact_name = corporation.corporation_name
            corporation_id = corporation.corporation_id
            corporation_name = corporation.corporation_name
            corporation_ticker = corporation.ticker
            alliance_id = None
            alliance_name = ""
            has_scopes = (
                not corporation.is_npc
                and corporation.user_has_all_member_tokens(
                    user=req.user, quick_check=quick_check
                )
            )
        else:
            contact_name = ""
            contact_icon_url = ""
            corporation_id = None
            corporation_name = ""
            corporation_ticker = ""
            alliance_id = None
            alliance_name = ""
            has_scopes = False

        if req.is_standing_revocation:
            reason = req.get_reason_display()
        else:
            reason = None
        try:
            my_contact = contacts[req.contact_id]
        except KeyError:
            labels = []
        else:
            labels = [obj.name for obj in my_contact.labels.all()]
        requests_data.append(
            {
                "contact_id": req.contact_id,
                "contact_name": contact_name,
                "contact_icon_url": contact_icon_url,
                "corporation_id": corporation_id,
                "corporation_name": corporation_name,
                "corporation_ticker": corporation_ticker,
                "alliance_id": alliance_id,
                "alliance_name": alliance_name,
                "request_date": req.request_date.isoformat(),
                "action_date": req.action_date.isoformat() if req.action_date else None,
                "has_scopes": has_scopes,
                "state": state_name,
                "reason": reason,
                "labels": sorted(labels),
                "main_character_name": main_character_name,
                "main_character_ticker": main_character_ticker,
                "main_character_icon_url": main_character_icon_url,
                "actioned": req.is_actioned,
                "is_effective": req.is_effective,
                "is_corporation": req.is_corporation,
                "is_character": req.is_character,
                "action_by": req.action_by.username if req.action_by else "(System)",
            }
        )
    return requests_data


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_requests_write(request, contact_id):
    contact_id = int(contact_id)
    logger.debug("manage_requests_write called by %s", request.user)
    if request.method == "PUT":
        actioned = 0
        for r in StandingRequest.objects.filter(contact_id=contact_id):
            r.mark_actioned(request.user)
            RequestLogEntry.objects.create_from_standing_request(
                r, RequestLogEntry.Action.CONFIRMED, request.user
            )
            actioned += 1
        if actioned > 0:
            return HttpResponseNoContent()
        return HttpResponseNotFound()
    elif request.method == "DELETE":
        standing_request = get_object_or_404(StandingRequest, contact_id=contact_id)
        RequestLogEntry.objects.create_from_standing_request(
            standing_request, RequestLogEntry.Action.REJECTED, request.user
        )
        standing_request.delete()
        if SR_NOTIFICATIONS_ENABLED:
            entity_name = EveEntity.objects.resolve_name(contact_id)
            title = _("Standing request for %s rejected" % entity_name)
            message = _(
                "Your standing request for '%s' has been rejected by %s."
                % (entity_name, request.user)
            )
            notify(user=standing_request.user, title=title, message=message)

        return HttpResponseNoContent()
    return HttpResponseNotFound()


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_revocations_write(request, contact_id):
    contact_id = int(contact_id)
    logger.debug(
        "manage_revocations_write called by %s for contact_id %s",
        str(request.user),
        contact_id,
    )
    if request.method == "PUT":
        actioned = 0
        for r in StandingRevocation.objects.filter(
            contact_id=contact_id, action_date__isnull=True
        ):
            r.mark_actioned(request.user)
            RequestLogEntry.objects.create_from_standing_request(
                r, RequestLogEntry.Action.CONFIRMED, request.user
            )
            actioned += 1
        if actioned > 0:
            return HttpResponseNoContent()
        return HttpResponseNotFound
    elif request.method == "DELETE":
        standing_revocations_qs = StandingRevocation.objects.filter(
            contact_id=contact_id
        )
        standing_revocation = standing_revocations_qs.first()
        RequestLogEntry.objects.create_from_standing_request(
            standing_revocation, RequestLogEntry.Action.REJECTED, request.user
        )
        standing_revocations_qs.delete()
        if SR_NOTIFICATIONS_ENABLED and standing_revocation.user:
            entity_name = EveEntity.objects.resolve_name(contact_id)
            title = _("Standing revocation for %s rejected" % entity_name)
            message = _(
                "Your standing revocation for '%s' "
                "has been rejected by %s." % (entity_name, request.user)
            )
            notify(user=standing_revocation.user, title=title, message=message)
        return HttpResponseNoContent()
    return HttpResponseNotFound()


###################
# View requests #
###################


@login_required
@permission_required("standingsrequests.affect_standings")
def view_active_requests(request):
    context = {
        "organization": BaseConfig.standings_source_entity(),
        "requests_count": _standing_requests_to_view().count(),
    }
    return render(
        request, "standingsrequests/requests.html", add_common_context(request, context)
    )


@login_required
@permission_required("standingsrequests.affect_standings")
def view_requests_json(request):

    response_data = _compose_standing_requests_data(
        _standing_requests_to_view(), quick_check=True
    )
    return JsonResponse(response_data, safe=False)


def _standing_requests_to_view() -> models.QuerySet:
    return (
        StandingRequest.objects.filter(is_effective=True)
        .select_related("user__profile")
        .order_by("-request_date")
    )



from datetime import timedelta

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from esi.models import Token
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import CharacterOwnership, State
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from app_utils.helpers import default_if_none
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import SR_REQUIRED_SCOPES, SR_STANDING_TIMEOUT_HOURS
from .constants import OperationMode
from .core import BaseConfig, ContactType, MainOrganizations
from .helpers.evecorporation import EveCorporation
from .managers import (
    AbstractStandingsRequestManager,
    CharacterAffiliationManager,
    ContactQuerySet,
    ContactSetManager,
    CorporationDetailsManager,
    FrozenAltManager,
    FrozenAuthUserManager,
    RequestLogEntryManager,
    StandingRequestManager,
    StandingRevocationManager,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def get_or_create_sentinel_user() -> User:
    """Get or create the sentinel user."""
    return User.objects.get_or_create(username="deleted")[0]


class ContactSet(models.Model):
    """Set of contacts from configured alliance or corporation
    which defines its current standings
    """

    date = models.DateTimeField(auto_now_add=True, db_index=True)
    name = models.CharField(max_length=254)

    objects = ContactSetManager()

    class Meta:
        get_latest_by = "date"
        permissions = (
            ("view", "User can view standings"),
            ("download", "User can export standings to a CSV file"),
        )

    def __str__(self):
        return str(self.date)

    def __repr__(self):
        return f"{type(self).__name__}(pk={self.pk}, date='{self.date}')"

    def contact_has_satisfied_standing(self, contact_id: int) -> bool:
        """Return True if give contact has standing exists"""
        try:
            contact = self.contacts.get(eve_entity_id=contact_id)
        except Contact.DoesNotExist:
            return False
        return contact.is_standing_satisfied

    def generate_standing_requests_for_blue_alts(self) -> int:
        """Automatically creates effective standings requests for
        alt characters on Auth that already have blue standing in-game.

        return count of generated standings requests
        """
        logger.info("Started generating standings request for blue alts.")
        owned_characters_qs = EveCharacter.objects.filter(
            character_ownership__isnull=False
        )
        created_counter = 0
        for alt in owned_characters_qs:
            user = alt.character_ownership.user
            if (
                not MainOrganizations.is_character_a_member(alt)
                and not StandingRequest.objects.filter(
                    user=user, contact_id=alt.character_id
                ).exists()
                and not StandingRevocation.objects.filter(
                    contact_id=alt.character_id
                ).exists()
                and self.contact_has_satisfied_standing(alt.character_id)
            ):
                sr = StandingRequest.objects.get_or_create_2(
                    user=user,
                    contact_id=alt.character_id,
                    contact_type=StandingRequest.ContactType.CHARACTER,
                )
                sr.mark_actioned(user=None, reason=sr.Reason.STANDING_IN_GAME)
                sr.mark_effective()
                RequestLogEntry.objects.create_from_standing_request(
                    sr, RequestLogEntry.Action.CONFIRMED, None
                )
                logger.info(
                    "Generated standings request for blue alt %s "
                    "belonging to user %s.",
                    alt,
                    user,
                )
                created_counter += 1

        logger.info(
            "Completed generating %d standings request for blue alts.",
            created_counter,
        )
        return created_counter

    @staticmethod
    def required_esi_scope() -> str:
        """returns the required ESI scopes for syncing"""
        if BaseConfig.operation_mode is OperationMode.ALLIANCE:
            return "esi-alliances.read_contacts.v1"
        elif BaseConfig.operation_mode is OperationMode.CORPORATON:
            return "esi-corporations.read_contacts.v1"
        else:
            raise NotImplementedError()


class ContactLabel(models.Model):
    """A contact label"""

    contact_set = models.ForeignKey(
        ContactSet, on_delete=models.CASCADE, related_name="labels"
    )
    label_id = models.BigIntegerField(db_index=True)
    name = models.CharField(max_length=254, db_index=True)

    def __str__(self):
        return self.name

    def __repr__(self):
        return (
            f"{type(self).__name__}(pk={self.pk}, "
            f"label_id={self.label_id}, name='{self.name}')"
        )


class Contact(models.Model):
    """An Eve Online contact."""

    contact_set = models.ForeignKey(
        ContactSet, on_delete=models.CASCADE, related_name="contacts"
    )
    eve_entity = models.ForeignKey(
        EveEntity, on_delete=models.CASCADE, related_name="standingrequests_contact"
    )
    standing = models.FloatField(db_index=True)
    labels = models.ManyToManyField(ContactLabel, related_name="contacts")
    is_watched = models.BooleanField(default=False)

    objects = ContactQuerySet.as_manager()

    def __str__(self):
        return self.eve_entity.name

    def __repr__(self):
        return (
            f"{type(self).__name__}(pk={self.pk}, "
            f"contact_id={self.eve_entity_id}, name='{self.eve_entity.name}', "
            f"standing={self.standing})"
        )

    @property
    def name(self) -> str:
        return self.eve_entity.name

    @property
    def is_standing_satisfied(self) -> str:
        return StandingRequest.is_standing_satisfied(self.standing)


class AbstractStandingsRequest(models.Model):
    """Base class for a standing request"""

    class ContactType(models.TextChoices):
        """Possible contact types to make a request for."""

        CHARACTER = "character", _("character")
        CORPORATION = "corporation", _("corporation")

    class Reason(models.TextChoices):
        """Reason for requesting or revoking a standing."""

        NONE = "NO", _("None recorded")
        OWNER_REQUEST = "OR", _("Requested by character owner")
        LOST_PERMISSION = "LP", _("Character owner has lost permission")
        MISSING_CORP_TOKEN = "CT", _("Not all corp tokens are recorded in Auth.")
        REVOKED_IN_GAME = "RG", _("Standing has been revoked in game")
        STANDING_IN_GAME = "SG", _("Already has standing in game")

    # Standing less than or equal
    EXPECT_STANDING_LTEQ = 10.0

    # Standing greater than or equal
    EXPECT_STANDING_GTEQ = -10.0

    # permission needed to request standing
    REQUEST_PERMISSION_NAME = "standingsrequests.request_standings"

    contact_id = models.PositiveIntegerField(
        db_index=True, help_text="EVE Online ID of contact this standing is for"
    )
    contact_type_id = models.PositiveIntegerField(
        db_index=True, help_text="EVE Online Type ID of this contact"
    )
    request_date = models.DateTimeField(
        auto_now_add=True, db_index=True, help_text="datetime this request was created"
    )
    action_by = models.ForeignKey(
        User,
        default=None,
        null=True,
        on_delete=models.SET_DEFAULT,
        db_index=True,
        help_text="standing manager that accepted or rejected this requests",
    )
    action_date = models.DateTimeField(
        null=True, db_index=True, help_text="datetime of action by standing manager"
    )
    is_effective = models.BooleanField(
        default=False,
        db_index=True,
        help_text="True, when this standing is also set in-game, else False",
    )
    effective_date = models.DateTimeField(
        null=True, help_text="Datetime when this standing was set active in-game"
    )

    objects = AbstractStandingsRequestManager()

    class Meta:
        permissions = (
            ("affect_standings", "User can process standings requests."),
            ("request_standings", "User can request standings."),
        )

    def __repr__(self) -> str:
        try:
            user_str = f", user='{self.user}'"
        except AttributeError:
            user_str = ""

        return (
            f"{type(self).__name__}(pk={self.pk}, contact_id={self.contact_id}"
            f"{user_str}, is_effective={self.is_effective})"
        )

    @property
    def is_character(self) -> bool:
        return ContactType.is_character(self.contact_type_id)

    @property
    def is_corporation(self) -> bool:
        return ContactType.is_corporation(self.contact_type_id)

    @property
    def is_actioned(self) -> bool:
        return self.action_date is not None and not self.is_effective

    @property
    def is_pending(self) -> bool:
        return self.action_date is None and self.is_effective is False

    @property
    def is_standing_request(self) -> bool:
        return type(self) is StandingRequest

    @property
    def is_standing_revocation(self) -> bool:
        return type(self) is StandingRevocation

    @classmethod
    def is_standing_satisfied(cls, standing: float) -> bool:
        if standing is not None:
            return (
                cls.EXPECT_STANDING_GTEQ <= float(standing) <= cls.EXPECT_STANDING_LTEQ
            )
        else:
            return False

    @classmethod
    def contact_type_2_id(cls, contact_type) -> int:
        if contact_type == cls.ContactType.CHARACTER:
            return ContactType.character_id
        elif contact_type == cls.ContactType.CORPORATION:
            return ContactType.corporation_id
        else:
            raise ValueError("Invalid contact type")

    @classmethod
    def contact_id_2_type(cls, contact_type_id) -> str:
        if contact_type_id in ContactType.character_ids:
            return cls.ContactType.CHARACTER.value
        elif contact_type_id in ContactType.corporation_ids:
            return cls.ContactType.CORPORATION.value
        raise ValueError("Invalid contact type")

    def evaluate_effective_standing(self, check_only: bool = False) -> bool:
        """
        Check and mark a standing as satisfied
        :param check_only: Check the standing only, take no action
        """
        try:
            logger.debug("Checking standing for %d", self.contact_id)
            latest = ContactSet.objects.latest()
            contact = latest.contacts.get(eve_entity_id=self.contact_id)
            if self.is_standing_satisfied(contact.standing):
                # Standing is satisfied
                logger.debug("Standing satisfied for %d", self.contact_id)
                if not check_only:
                    self.mark_effective()
                return True

        except ObjectDoesNotExist:
            logger.debug(
                "No standing set for %d, checking if neutral is OK", self.contact_id
            )
            if self.is_standing_satisfied(0):
                # Standing satisfied but deleted (neutral)
                logger.debug(
                    "Standing satisfied but deleted (neutral) for %d", self.contact_id
                )
                if not check_only:
                    self.mark_effective()
                return True

        # Standing not satisfied
        logger.debug("Standing NOT satisfied for %d", self.contact_id)
        return False

    def mark_effective(self, date=None):
        """
        Marks a standing as effective (standing exists in game)
        from the current or supplied TZ aware datetime
        :param date: TZ aware datetime object of when the standing became effective
        :return:
        """
        logger.debug("Marking standing for %d as effective", self.contact_id)
        self.is_effective = True
        self.effective_date = date if date else now()
        self.save()

    def mark_actioned(self, user, date=None, reason=None):
        """
        Marks a standing as actioned (user has made the change in game)
        with the current or supplied TZ aware datetime
        :param user: Actioned By django User
        :param date: TZ aware datetime object of when the action was taken
        :return:
        """
        logger.debug("Marking standing for %d as actioned", self.contact_id)
        self.action_by = user
        self.action_date = date if date else now()
        if reason:
            self.reason = reason
        self.save()

    def check_actioned_timeout(self):
        """
        Check that a standing hasn't been marked as actioned
        and is still not effective ~24hr later
        :return: User if the actioned has timed out, False if it has not,
        None if the check was unsuccessful
        """
        logger.debug("Checking standings request timeout")
        if self.is_effective:
            logger.debug("Standing is already marked as effective...")
            return None

        if self.action_by is None:
            logger.debug("Standing was never actioned, cannot timeout")
            return None

        try:
            latest = ContactSet.objects.latest()
        except ContactSet.DoesNotExist:
            logger.debug("Cannot check standing timeout, no standings available")
            return None

        # Reset request that has not become effective after timeout expired
        if self.action_date + timedelta(hours=SR_STANDING_TIMEOUT_HOURS) < latest.date:
            logger.info(
                "Standing actioned timed out, resetting actioned for contact_id %d",
                self.contact_id,
            )
            actioner = self.action_by
            self.action_by = None
            self.action_date = None
            self.save()
            return actioner
        return False

    def reset_to_initial(self) -> None:
        """
        Reset a standing back to its initial creation state
        (Not actioned and not effective)
        :return:
        """
        self.is_effective = False
        self.effective_date = None
        self.action_by = None
        self.action_date = None
        self.save()


class StandingRequest(AbstractStandingsRequest):
    """A change request to get standing for a character or corporation

    OR a record representing that a character or corporation currently has standing

    Standing Requests (SR) can have one of 3 states:
    - new: Newly created SRs represent a new request from a user. They are not actioned and not effective
    - actionied: A standing manager marks a SR as actionied, once he has set the new standing in-game
    - effective: Once the new standing is returned from the API a SR is marked effective. Effective SRs stay in database to represent that a user has standing.
    """

    EXPECT_STANDING_GTEQ = 0.01

    reason = models.CharField(
        max_length=2,
        choices=AbstractStandingsRequest.Reason.choices,
        default=AbstractStandingsRequest.Reason.NONE,
    )
    user = models.ForeignKey(User, on_delete=models.CASCADE)

    objects = StandingRequestManager()

    def remove(self):
        """Remove this standing request."""
        if self.is_character:
            return self._remove_character_standing()
        elif self.is_corporation:
            return self._remove_corporation_request()
        raise NotImplementedError()

    def _remove_character_standing(self) -> bool:
        """Remove effective character standing for user if possible."""
        try:
            character = EveCharacter.objects.get(character_id=self.contact_id)
        except EveCharacter.DoesNotExist:
            return False
        if MainOrganizations.is_character_a_member(character):
            logger.warning(
                "%s: Character %s of user %s is in organization. Can not remove standing",
                self,
                character,
                self.user,
            )
            return False
        if StandingRevocation.objects.has_pending_request(self.contact_id):
            logger.debug(
                "%s: User %s already has a pending standing revocation for character %d",
                self,
                self.user,
                self.contact_id,
            )
            return False
        self.delete(reason=StandingRevocation.Reason.OWNER_REQUEST)
        return True

    def _remove_corporation_request(self) -> bool:
        """Remove effective corporation standing and pending requests
        for user if possible.
        """
        try:
            contact_set = ContactSet.objects.latest()
        except ContactSet.DoesNotExist:
            logger.warning("Failed to get a contact set")
            return False
        if (
            self.is_pending or self.is_actioned
        ) and not StandingRevocation.objects.has_pending_request(self.contact_id):
            logger.debug(
                "%s: Removing standings requests by user %s",
                self,
                self.user,
            )
            self.delete(reason=StandingRevocation.Reason.OWNER_REQUEST)
            return True
        if not contact_set.contact_has_satisfied_standing(self.contact_id):
            logger.debug("%s: Can not remove standing - no standings exist", self)
            return False
        # Manual revocation required
        logger.debug("%s: Creating standings revocation by user %s", self, self.user)
        StandingRevocation.objects.add_revocation(
            contact_id=self.contact_id,
            contact_type=StandingRevocation.ContactType.CORPORATION,
            user=self.user,
            reason=StandingRevocation.Reason.OWNER_REQUEST,
        )
        return True

    def delete(self, using=None, keep_parents=False, reason=None):
        """
        Add a revocation before deleting if the standing has been
        actioned (pending) or is effective and
        doesn't already have a pending revocation request.
        """
        if self.action_by is not None or self.is_effective:
            # Check if theres not already a revocation pending
            if not StandingRevocation.objects.has_pending_request(self.contact_id):
                logger.debug(
                    "Adding revocation for deleted request "
                    "with contact_id %d type %s",
                    self.contact_id,
                    self.contact_type_id,
                )
                StandingRevocation.objects.add_revocation(
                    contact_id=self.contact_id,
                    contact_type=self.contact_id_2_type(self.contact_type_id),
                    user=self.user,
                    reason=reason,
                )
            else:
                logger.debug(
                    "Revocation already pending for deleted request "
                    "with contact_id %d type %s",
                    self.contact_id,
                    self.contact_type_id,
                )
        else:
            logger.debug(
                "Standing never effective, no revocation required "
                "for deleted request with contact_id %d type %s",
                self.contact_id,
                self.contact_type_id,
            )

        logger.debug("%s: Removing standing request by user %s", self, self.user)
        super().delete(using, keep_parents)

    @classmethod
    def can_request_corporation_standing(cls, corporation_id: int, user: User) -> bool:
        """
        Checks if given user owns all of the required corp tokens for standings to be permitted

        Params
        - corporation_id: corp to check for
        - user: User to check for

        returns True if they can request standings, False if they cannot
        """
        corporation = EveCorporation.get_by_id(corporation_id)
        return (
            corporation is not None
            and not corporation.is_npc
            and corporation.user_has_all_member_tokens(user)
        )

    @classmethod
    def has_required_scopes_for_request(
        cls, character: EveCharacter, user: User = None, quick_check: bool = False
    ) -> bool:
        """returns true if given character has the required scopes
        for issueing a standings request else false

        Params:
        - user: provide User object to shorten processing time
        - quick: if True will not check if tokens are valid to save time
        """
        if not user:
            try:
                ownership = CharacterOwnership.objects.select_related(
                    "user", "user__profile__state"
                ).get(character__character_id=character.character_id)
            except CharacterOwnership.DoesNotExist:
                return False
            else:
                user = ownership.user
        try:
            state_name = user.profile.state.name
        except ObjectDoesNotExist:
            return False
        scopes_string = " ".join(cls.get_required_scopes_for_state(state_name))
        token_qs = Token.objects.filter(
            character_id=character.character_id
        ).require_scopes(scopes_string)
        if not quick_check:
            token_qs = token_qs.require_valid()
        return token_qs.exists()

    @staticmethod
    def get_required_scopes_for_state(state_name: str) -> list:
        state_name = "" if not state_name else state_name
        return (
            SR_REQUIRED_SCOPES[state_name]
            if state_name in SR_REQUIRED_SCOPES
            else list()
        )


class StandingRevocation(AbstractStandingsRequest):
    """A standing revocation"""

    EXPECT_STANDING_LTEQ = 0.0

    reason = models.CharField(
        max_length=2,
        choices=AbstractStandingsRequest.Reason.choices,
        default=AbstractStandingsRequest.Reason.NONE,
    )
    user = models.ForeignKey(
        User, on_delete=models.SET_DEFAULT, default=None, null=True
    )

    objects = StandingRevocationManager()


class CharacterAffiliation(models.Model):
    """Affiliation of a character."""

    character = models.OneToOneField(
        EveEntity,
        on_delete=models.CASCADE,
        primary_key=True,
        related_name="character_affiliation",
    )
    corporation = models.ForeignKey(
        EveEntity, on_delete=models.CASCADE, related_name="+"
    )
    alliance = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        related_name="+",
    )
    faction = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        related_name="+",
    )
    eve_character = models.ForeignKey(
        EveCharacter,
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        related_name="+",
        help_text="Related auth character (if any)",
    )
    updated = models.DateTimeField(auto_now_add=True)

    objects = CharacterAffiliationManager()

    def __str__(self) -> str:
        return self.character.name

    @cached_property
    def character_name(self) -> str:
        """Return character name for main."""
        return self.character.name if self.character.name else None

    def entity_ids(self) -> set:
        """Ids of all entities."""
        return set(
            filter(
                lambda x: x is not None,
                [
                    self.character_id,
                    self.corporation_id,
                    self.alliance_id,
                    self.faction_id,
                ],
            )
        )


class CorporationDetails(models.Model):
    """A corporation affiliation."""

    corporation = models.OneToOneField(
        EveEntity,
        on_delete=models.CASCADE,
        primary_key=True,
        related_name="corporation_details",
    )

    alliance = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        related_name="+",
    )
    ceo = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        related_name="+",
    )
    faction = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        related_name="+",
    )
    member_count = models.PositiveIntegerField()
    ticker = models.CharField(max_length=255)

    objects = CorporationDetailsManager()

    def __str__(self) -> str:
        return self.corporation.name


class FrozenModelMixin:
    """Objects of this model type can only be created, but not updated."""

    def save(self, *args, **kwargs) -> None:
        if self.pk is None:
            super().save(*args, **kwargs)
        else:
            raise RuntimeError("No updates allowed for this object.")


class RequestLogEntry(FrozenModelMixin, models.Model):
    class Action(models.TextChoices):
        CONFIRMED = "CN", _("confirmed")
        REJECTED = "RJ", _("rejected")

    class RequestType(models.TextChoices):
        REQUEST = "RQ", _("request")
        REVOCATION = "RV", _("revocation")

    action = models.CharField(max_length=2, choices=Action.choices)
    action_by = models.ForeignKey(
        "FrozenAuthUser",
        on_delete=models.CASCADE,
        null=True,
        help_text=(
            "Main who performed the action. "
            "None means the action was performed automatically by the app."
        ),
    )
    created_at = models.DateTimeField(auto_now=True)
    reason = models.CharField(
        max_length=2, choices=AbstractStandingsRequest.Reason.choices
    )
    request_type = models.CharField(max_length=2, choices=RequestType.choices)
    requested_at = models.DateTimeField()
    requested_by = models.ForeignKey(
        "FrozenAuthUser", on_delete=models.CASCADE, related_name="+"
    )
    requested_for = models.ForeignKey(
        "FrozenAlt",
        on_delete=models.CASCADE,
        help_text="Alt character or corporation to change standing for",
    )

    objects = RequestLogEntryManager()

    class Meta:
        verbose_name = "request log"
        verbose_name_plural = "request log"

    def __str__(self) -> str:
        return f"{self.created_at}-{self.action}"


class FrozenAuthUser(FrozenModelMixin, models.Model):
    """Main with user, character and affiliations.
    Objects are frozen at creation and can not be changed.
    """

    alliance = models.ForeignKey(
        EveEntity, on_delete=models.SET_NULL, null=True, related_name="+"
    )
    character = models.ForeignKey(
        EveEntity, on_delete=models.SET_NULL, null=True, related_name="+"
    )
    corporation = models.ForeignKey(
        EveEntity, on_delete=models.SET_NULL, null=True, related_name="+"
    )
    faction = models.ForeignKey(
        EveEntity, on_delete=models.SET_NULL, null=True, related_name="+"
    )
    state = models.ForeignKey(
        State, on_delete=models.SET_NULL, null=True, related_name="+"
    )
    user = models.ForeignKey(
        User, on_delete=models.SET(get_or_create_sentinel_user), related_name="+"
    )

    objects = FrozenAuthUserManager()

    def __str__(self) -> str:
        return self.character.name if self.character else self.user.username

    def html(self) -> str:
        """Output as html."""
        if self.character:
            return format_html(
                "{}<br>{}",
                default_if_none(self.character, "-"),
                default_if_none(self.corporation, "-"),
            )
        return str(self.user)


class FrozenAlt(FrozenModelMixin, models.Model):
    """Alt with alignmants. Objects are frozen at creation and can not be changed."""

    class Category(models.TextChoices):
        CHARACTER = "CH", "character"
        CORPORATION = "CP", "corporation"

    alliance = models.ForeignKey(
        EveEntity, on_delete=models.SET_NULL, null=True, related_name="+"
    )
    character = models.ForeignKey(
        EveEntity, on_delete=models.SET_NULL, null=True, related_name="+"
    )
    corporation = models.ForeignKey(
        EveEntity, on_delete=models.SET_NULL, null=True, related_name="+"
    )
    category = models.CharField(max_length=2, choices=Category.choices)
    faction = models.ForeignKey(
        EveEntity, on_delete=models.SET_NULL, null=True, related_name="+"
    )

    objects = FrozenAltManager()

    def __str__(self) -> str:
        return str(self.character) if self.character else (self.corporation)

    @property
    def is_character(self) -> bool:
        return self.category == self.Category.CHARACTER

    @property
    def is_corporation(self) -> bool:
        return self.category == self.Category.CORPORATION

    def html(self) -> str:
        """Output as html."""
        if self.is_character:
            return format_html(
                "{}<br>{}",
                default_if_none(self.character, "-"),
                default_if_none(self.corporation, "-"),
            )
        return str(self.corporation)



# Generated by Django 3.2.9 on 2021-12-22 16:37

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import standingsrequests.models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("authentication", "0019_merge_20211026_0919"),
        ("eveuniverse", "0005_type_materials_and_sections"),
        ("standingsrequests", "0009_make_ceo_optional"),
    ]

    operations = [
        migrations.CreateModel(
            name="FrozenAlt",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "category",
                    models.CharField(
                        choices=[("CH", "character"), ("CP", "corporation")],
                        max_length=2,
                    ),
                ),
                (
                    "alliance",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "faction",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
            ],
            bases=(standingsrequests.models.FrozenModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name="FrozenAuthUser",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "alliance",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "faction",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "state",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="authentication.state",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=models.SET(
                            standingsrequests.models.get_or_create_sentinel_user
                        ),
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            bases=(standingsrequests.models.FrozenModelMixin, models.Model),
        ),
        migrations.AddField(
            model_name="standingrequest",
            name="reason",
            field=models.CharField(
                choices=[
                    ("NO", "None recorded"),
                    ("OR", "Requested by character owner"),
                    ("LP", "Character owner has lost permission"),
                    ("CT", "Not all corp tokens are recorded in Auth."),
                    ("RG", "Standing has been revoked in game"),
                    ("SG", "Already has standing in game"),
                ],
                default="NO",
                max_length=2,
            ),
        ),
        migrations.AlterField(
            model_name="standingrevocation",
            name="reason",
            field=models.CharField(
                choices=[
                    ("NO", "None recorded"),
                    ("OR", "Requested by character owner"),
                    ("LP", "Character owner has lost permission"),
                    ("CT", "Not all corp tokens are recorded in Auth."),
                    ("RG", "Standing has been revoked in game"),
                    ("SG", "Already has standing in game"),
                ],
                default="NO",
                max_length=2,
            ),
        ),
        migrations.CreateModel(
            name="RequestLogEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "action",
                    models.CharField(
                        choices=[("CN", "confirmed"), ("RJ", "rejected")], max_length=2
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now=True)),
                (
                    "reason",
                    models.CharField(
                        choices=[
                            ("NO", "None recorded"),
                            ("OR", "Requested by character owner"),
                            ("LP", "Character owner has lost permission"),
                            ("CT", "Not all corp tokens are recorded in Auth."),
                            ("RG", "Standing has been revoked in game"),
                            ("SG", "Already has standing in game"),
                        ],
                        max_length=2,
                    ),
                ),
                (
                    "request_type",
                    models.CharField(
                        choices=[("RQ", "request"), ("RV", "revocation")], max_length=2
                    ),
                ),
                ("requested_at", models.DateTimeField()),
                (
                    "action_by",
                    models.ForeignKey(
                        help_text="Main who performed the action. None means the action was performed automatically by the app.",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="standingsrequests.frozenauthuser",
                    ),
                ),
                (
                    "requested_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="standingsrequests.frozenauthuser",
                    ),
                ),
                (
                    "requested_for",
                    models.ForeignKey(
                        help_text="Alt character or corporation to change standing for",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="standingsrequests.frozenalt",
                    ),
                ),
            ],
            options={
                "verbose_name": "request log",
                "verbose_name_plural": "request log",
            },
            bases=(standingsrequests.models.FrozenModelMixin, models.Model),
        ),
    ]



from enum import IntEnum

from django.db.models import TextChoices


class OperationMode(TextChoices):
    ALLIANCE = "alliance"
    CORPORATON = "corporation"


class CreateCharacterRequestError(IntEnum):
    NO_ERROR = 0
    USER_IS_NOT_OWNER = 1
    CHARACTER_HAS_REQUEST = 2
    CHARACTER_IS_MISSING_SCOPES = 3
    UNKNOWN_ERROR = 99



from enum import IntEnum

from django.utils.functional import classproperty
from eveuniverse.models import EveEntity

from allianceauth.eveonline.models import EveCharacter

from .app_settings import (
    SR_OPERATION_MODE,
    STANDINGS_API_CHARID,
    STR_ALLIANCE_IDS,
    STR_CORP_IDS,
)
from .constants import OperationMode


class MainOrganizations:
    """Configured main alliances and corporations from settings"""

    @classmethod
    def is_character_a_member(cls, character: EveCharacter) -> bool:
        """Check if the Pilot is in the auth instances organisation

        character: EveCharacter

        returns True if the character is in the organisation, False otherwise
        """
        return (
            character.corporation_id in cls.corporation_ids
            or character.alliance_id in cls.alliance_ids
        )

    @classproperty
    def corporation_ids(cls) -> set:
        return {int(org_id) for org_id in list(STR_CORP_IDS)}

    @classproperty
    def alliance_ids(cls) -> set:
        return {int(org_id) for org_id in list(STR_ALLIANCE_IDS)}


class BaseConfig:
    @classproperty
    def owner_character_id(cls) -> int:
        return STANDINGS_API_CHARID

    @classproperty
    def operation_mode(cls) -> OperationMode:
        """Return current operation mode."""
        return OperationMode(SR_OPERATION_MODE)

    @staticmethod
    def owner_character() -> EveCharacter:
        """returns the configured standings character"""
        try:
            return EveCharacter.objects.get(character_id=STANDINGS_API_CHARID)
        except EveCharacter.DoesNotExist:
            return EveCharacter.objects.create_character(STANDINGS_API_CHARID)

    @classmethod
    def standings_source_entity(cls) -> object:
        """returns the entity that all standings are fetched from

        returns None when in alliance mode, but character has no alliance
        """
        character = cls.owner_character()
        if cls.operation_mode is OperationMode.ALLIANCE:
            if character.alliance_id:
                entity, _ = EveEntity.objects.get_or_create_esi(
                    id=character.alliance_id
                )
            else:
                entity = None
        elif cls.operation_mode is OperationMode.CORPORATON:
            entity, _ = EveEntity.objects.get_or_create_esi(id=character.corporation_id)
        else:
            raise NotImplementedError()

        return entity


class ContactType(IntEnum):
    CHARACTER_AMARR_TYPE_ID = 1373
    CHARACTER_NI_KUNNI_TYPE_ID = 1374
    CHARACTER_CIVRE_TYPE_ID = 1375
    CHARACTER_DETEIS_TYPE_ID = 1376
    CHARACTER_GALLENTE_TYPE_ID = 1377
    CHARACTER_INTAKI_TYPE_ID = 1378
    CHARACTER_SEBIESTOR_TYPE_ID = 1379
    CHARACTER_BRUTOR_TYPE_ID = 1380
    CHARACTER_STATIC_TYPE_ID = 1381
    CHARACTER_MODIFIER_TYPE_ID = 1382
    CHARACTER_ACHURA_TYPE_ID = 1383
    CHARACTER_JIN_MEI_TYPE_ID = 1384
    CHARACTER_KHANID_TYPE_ID = 1385
    CHARACTER_VHEROKIOR_TYPE_ID = 1386
    CHARACTER_DRIFTER_TYPE_ID = 34574
    ALLIANCE_TYPE_ID = 16159
    CORPORATION_TYPE_ID = 2

    @classproperty
    def character_id(cls):
        return cls.CHARACTER_AMARR_TYPE_ID

    @classproperty
    def character_ids(cls):
        return {
            cls.CHARACTER_AMARR_TYPE_ID,
            cls.CHARACTER_NI_KUNNI_TYPE_ID,
            cls.CHARACTER_CIVRE_TYPE_ID,
            cls.CHARACTER_DETEIS_TYPE_ID,
            cls.CHARACTER_GALLENTE_TYPE_ID,
            cls.CHARACTER_INTAKI_TYPE_ID,
            cls.CHARACTER_SEBIESTOR_TYPE_ID,
            cls.CHARACTER_BRUTOR_TYPE_ID,
            cls.CHARACTER_STATIC_TYPE_ID,
            cls.CHARACTER_MODIFIER_TYPE_ID,
            cls.CHARACTER_ACHURA_TYPE_ID,
            cls.CHARACTER_JIN_MEI_TYPE_ID,
            cls.CHARACTER_KHANID_TYPE_ID,
            cls.CHARACTER_VHEROKIOR_TYPE_ID,
            cls.CHARACTER_DRIFTER_TYPE_ID,
        }

    @classproperty
    def corporation_ids(cls):
        return {cls.CORPORATION_TYPE_ID}

    @classproperty
    def corporation_id(cls):
        return cls.CORPORATION_TYPE_ID

    @classproperty
    def alliance_ids(cls):
        return {cls.ALLIANCE_TYPE_ID}

    @classproperty
    def alliance_id(cls):
        return cls.ALLIANCE_TYPE_ID

    @classmethod
    def is_character(cls, type_id):
        return type_id in cls.character_ids

    @classmethod
    def is_corporation(cls, type_id):
        return type_id in cls.corporation_ids

    @classmethod
    def is_alliance(cls, type_id):
        return type_id in cls.alliance_ids
